"""
Manage GPG keychains
====================

.. versionadded:: 2016.3.0

"""

import logging
from datetime import date

import salt.utils.dictupdate
import salt.utils.immutabletypes as immutabletypes
from salt.exceptions import CommandExecutionError, SaltInvocationError

log = logging.getLogger(__name__)

TRUST_MAP = immutabletypes.freeze(
    {
        "expired": "Expired",
        "unknown": "Unknown",
        "not_trusted": "Not Trusted",
        "marginally": "Marginally",
        "fully": "Fully Trusted",
        "ultimately": "Ultimately Trusted",
    }
)


class KeyNotContained(CommandExecutionError):
    """
    Raised when a data source does not contain a requested key
    """


def _expired_subkeys(key, max_days=100):
    """
    From a single key of the output of gpg.list_keys, iterate over
    subkeys and return those that have expired recently.
    We don't want to keep checking old subkeys for updates.
    """
    ret = {}
    if not max_days:
        return ret
    for subkey in key.get("subkeys", []):
        if subkey.get("expired"):
            # Only try to refresh subkeys that have expired recently
            expired_date = date.fromisoformat(subkey["expires"])
            if (date.today() - expired_date).days <= max_days:
                ret[subkey["keyid"]] = subkey
    return ret


def present(
    name,
    keys=None,
    user=None,
    keyserver=None,
    gnupghome=None,
    trust=None,
    keyring=None,
    source=None,
    skip_keyserver=False,
    text=None,
    subkey_maxage=100,
    **kwargs,
):
    """
    Ensure a GPG public key is present in the GPG keychain and
    that it is not expired.

    name
        The key ID of the GPG public key.

    keys
        The key ID or key IDs to add to the GPG keychain.

    user
        Add GPG keys to the specified user's keychain.

    keyserver
        The keyserver to retrieve the keys from.

    gnupghome
        Override GnuPG home directory.

    trust
        Trust level for the key in the keychain,
        ignored by default. Valid trust levels:
        expired, unknown, not_trusted, marginally,
        fully, ultimately

    keyring
        Limit the operation to this specific keyring, specified as
        a local filesystem path.

        .. versionadded:: 3007.0

    source
        A (list of) path(s)/URI to retrieve the key(s) from.
        By default, this works as a backup option in case retrieving a key
        from the keyserver fails.

        .. note::
            All listed sources will be iterated over in order until the first one found
            to contain the requested key. If multiple keys are managed in a single
            state, the effective sources are allowed to differ between keys.

        .. important::
            Internally, this uses :py:func:`gpg.read_key <salt.modules.gpg.read_key>`
            to list keys in the sources. If a source is not a keyring, on GnuPG <2.1,
            this can lead to unintentional decryption.

        .. versionadded:: 3008.0

    skip_keyserver
        Do not attempt to retrieve the key from the keyserver, only use ``source``.
        Irrelevant when ``text`` is passed. Defaults to false.

        .. versionadded:: 3008.0

    text
        Instead of retrieving the key(s) to import from a keyserver/URI,
        import them from this (armored) string.

        .. note::
            ``name`` or ``keys`` must still specify the expected key ID(s),
            so this cannot be used to indiscriminately import a keyring.
            Requires python-gnupg v0.5.1.

        .. versionadded:: 3008.0

    subkey_maxage
        If the managed key has expired subkeys, this state attempts an update.
        Since sometimes keys have long expired subkeys, it filters eligible subkeys
        that trigger the update check.
        This parameter specifies the maximum number of days since a subkey's
        expiration for the key to be eligible. Defaults to ``100``, meaning
        subkeys that have expired more than 100 days ago do not trigger an attempt.
        Set this to a falsy value to skip the explicit management of subkeys.

        .. versionadded:: 3008.0
    """

    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    try:
        if not text and skip_keyserver and not source:
            raise SaltInvocationError(
                "When skipping keyservers, you must provide at least one source"
            )

        if trust and trust not in TRUST_MAP:
            raise SaltInvocationError(f"Invalid trust level {trust}")

        _current_keys = __salt__["gpg.list_keys"](
            user=user, gnupghome=gnupghome, keyring=keyring
        )
    except (CommandExecutionError, SaltInvocationError) as err:
        ret["result"] = False
        ret["comment"] = str(err)
        return ret

    current_keys = {}
    expired_keys = []
    expired_subkeys = {}
    for key in _current_keys:
        keyid = key["keyid"]
        current_keys[keyid] = {}
        current_keys[keyid]["trust"] = key["trust"]
        current_keys[keyid]["subkeys"] = key.get("subkeys", {})
        if key.get("expired"):
            expired_keys.append(keyid)
        key_expired_subs = _expired_subkeys(key, max_days=subkey_maxage)
        if key_expired_subs:
            expired_subkeys[keyid] = key_expired_subs

    if not keys:
        keys = name
    if isinstance(keys, str):
        keys = [keys]
    key_res = {}

    # First, ensure all keys are present
    for key in keys:
        key_res[key] = []
        try:
            is_expired = key in expired_keys
            has_expired_subkeys = key in expired_subkeys
            refresh = is_expired or has_expired_subkeys
            if key in current_keys and not refresh:
                key_res[key].append(f"GPG Public Key {key} already in keychain")
                continue
            if __opts__["test"]:
                ret["result"] = None
                if refresh:
                    if is_expired:
                        key_res[key] = [
                            f"Would have attempted to update {key} because it is expired"
                        ]
                    else:
                        key_res[key] = [
                            f"Would have attempted to update {key} because it has expired subkeys:"
                        ] + [
                            f"  - {subkey} (expired on {data['expires']})"
                            for subkey, data in expired_subkeys[key].items()
                        ]
                    salt.utils.dictupdate.set_dict_key_value(
                        ret, f"changes:{key}:refresh", True
                    )
                else:
                    key_res[key] = [f"Would have added {key} to GPG keychain"]
                    salt.utils.dictupdate.set_dict_key_value(
                        ret, f"changes:{key}:added", True
                    )
                    current_keys[key] = {"trust": "unknown"}
                continue
            result = {}
            if text:
                result = _import_data(key, refresh, user, gnupghome, keyring, text=text)
            else:
                if not skip_keyserver:
                    result = __salt__["gpg.receive_keys"](
                        keyserver=keyserver,
                        keys=key,
                        user=user,
                        gnupghome=gnupghome,
                        keyring=keyring,
                    )
                    if refresh and result["res"]:
                        # If we're refreshing and no updated key could be found,
                        # ensure we're failing here.
                        result["res"] = any(
                            "updated: new" in x for x in result["message"]
                        )
                        if not is_expired and not result["res"]:
                            # Don't fail if we're only updating expired subkeys though
                            result["res"] = None
                    result["message"] = "\n".join(result["message"])
                if (not result or not result["res"]) and source:
                    if not isinstance(source, list):
                        source = [source]
                    prev_msg = ""
                    if result:
                        prev_msg = result["message"] + "\n"
                    for src in source:
                        sfn = __salt__["cp.cache_file"](src)
                        if sfn:
                            log.debug("Found source: %s", src)
                            try:
                                result = _import_data(
                                    key, refresh, user, gnupghome, keyring, path=sfn
                                )
                                break
                            except KeyNotContained as err:
                                if "expired" in str(err):
                                    log.warning(
                                        "Found source %s contains key %s, but it's expired",
                                        src,
                                        key,
                                    )
                    else:
                        raise CommandExecutionError(
                            prev_msg
                            + f"none of the specified sources were found or contained the (unexpired) key {key}."
                        )

            if result["res"] is False:
                raise CommandExecutionError(result["message"])
            new_key = __salt__["gpg.get_key"](
                keyid=key, user=user, gnupghome=gnupghome, keyring=keyring
            )
            if not new_key:
                raise CommandExecutionError(
                    result["message"]
                    + f"\nThe new key {key} could not be retrieved though."
                )
            if not refresh:
                salt.utils.dictupdate.set_dict_key_value(
                    ret, f"changes:{key}:added", True
                )
            if new_key.get("expired"):
                raise CommandExecutionError(
                    result["message"] + f"\nThe new key {key} is expired though."
                )
            if refresh:
                added_subs = {
                    subkey["keyid"] for subkey in new_key.get("subkeys", {})
                } - {subkey["keyid"] for subkey in current_keys[key]["subkeys"]}
                updated_subs = set()
                if has_expired_subkeys:
                    after_expired_subs = _expired_subkeys(
                        new_key, max_days=subkey_maxage
                    )
                    updated_subs = set(expired_subkeys[key]) - set(after_expired_subs)

                if is_expired:
                    key_res[key].append(f"Updated {key} because it was expired")
                    salt.utils.dictupdate.set_dict_key_value(
                        ret, f"changes:{key}:refresh", True
                    )
                elif added_subs or updated_subs:
                    key_res[key].append(
                        f"Updated {key} because it had expired subkeys."
                    )
                    salt.utils.dictupdate.set_dict_key_value(
                        ret, f"changes:{key}:refresh", True
                    )
                    if added_subs:
                        salt.utils.dictupdate.set_dict_key_value(
                            ret, f"changes:{key}:subkeys:added", list(added_subs)
                        )
                    if updated_subs:
                        salt.utils.dictupdate.set_dict_key_value(
                            ret, f"changes:{key}:subkeys:extended", list(updated_subs)
                        )
                else:
                    key_res[key].append(
                        f"Attempted to update {key} because it has expired subkeys, but no new signatures or keys were found"
                    )
            else:
                key_res[key].append(f"Added {key} to GPG keychain")
            current_keys[key] = {"trust": new_key["trust"]}
        except (CommandExecutionError, SaltInvocationError) as err:
            ret["result"] = False
            if is_expired:
                key_res[key].append(
                    "Existing key is expired, tried to fetch updated one"
                )
            elif has_expired_subkeys:
                key_res[key].append(
                    "Existing key has expired subkeys, tried to refresh"
                )
            key_res[key].extend(str(err).splitlines())

    # Now all possible keys are present, manage their trust if requested
    if trust:
        for key in keys:
            if key not in current_keys:
                # This means the key was not present and could not be retrieved
                continue
            try:
                if current_keys[key]["trust"] == TRUST_MAP[trust]:
                    key_res[key].append(
                        f"GPG Public Key {key} already in correct trust state"
                    )
                    continue
                if __opts__["test"]:
                    ret["result"] = None
                    key_res[key].append(
                        f"Would have set trust level for {key} to {trust}"
                    )
                    salt.utils.dictupdate.set_dict_key_value(
                        ret, f"changes:{key}:trust", trust
                    )
                    continue
                result = __salt__["gpg.trust_key"](
                    keyid=key,
                    trust_level=trust,
                    user=user,
                    gnupghome=gnupghome,
                    keyring=keyring,
                )
                if result["res"] is False:
                    raise CommandExecutionError(result["message"])
                key_res[key].append(f"Set trust level for {key} to {trust}")
                salt.utils.dictupdate.set_dict_key_value(
                    ret, f"changes:{key}:trust", trust
                )
            except (CommandExecutionError, SaltInvocationError) as err:
                ret["result"] = False
                key_res[key].append(str(err))
    final_res = {
        key: "\n  * " + "\n  * ".join(msgs) for key, msgs in key_res.items() if msgs
    }
    ret["comment"] = "\n".join(f"Key {key}:{msg}" for key, msg in final_res.items())
    return ret


def absent(
    name,
    keys=None,
    user=None,
    gnupghome=None,
    keyring=None,
    keyring_absent_if_empty=False,
    **kwargs,
):
    """
    Ensure a GPG public key is absent from the keychain.

    name
        The key ID of the GPG public key.

    keys
        The key ID or key IDs to remove from the GPG keychain.

    user
        Remove GPG keys from the specified user's keychain.

    gnupghome
        Override GnuPG home directory.

    keyring
        Limit the operation to this specific keyring, specified as
        a local filesystem path.

        .. versionadded:: 3007.0

    keyring_absent_if_empty
        Make sure to not leave behind an empty keyring file
        if ``keyring`` was specified. Defaults to false.

        .. versionadded:: 3007.0
    """

    ret = {"name": name, "result": True, "changes": {}, "comment": []}

    _current_keys = __salt__["gpg.list_keys"](
        user=user, gnupghome=gnupghome, keyring=keyring
    )

    current_keys = []
    for key in _current_keys:
        current_keys.append(key["keyid"])

    if not keys:
        keys = name

    if isinstance(keys, str):
        keys = [keys]

    for key in keys:
        if key in current_keys:
            if __opts__["test"]:
                ret["result"] = None
                ret["comment"].append(f"Would have deleted {key} from GPG keychain")
                salt.utils.dictupdate.append_dict_key_value(ret, "changes:deleted", key)
                continue
            result = __salt__["gpg.delete_key"](
                keyid=key,
                user=user,
                gnupghome=gnupghome,
                keyring=keyring,
            )
            if result["res"] is False:
                ret["result"] = result["res"]
                ret["comment"].append(result["message"])
            else:
                ret["comment"].append(f"Deleted {key} from GPG keychain")
                salt.utils.dictupdate.append_dict_key_value(ret, "changes:deleted", key)
        else:
            ret["comment"].append(f"{key} not found in GPG keychain")

    if __opts__["test"] or not ret["result"]:
        return ret

    _new_keys = [
        x["keyid"]
        for x in __salt__["gpg.list_keys"](
            user=user, gnupghome=gnupghome, keyring=keyring
        )
    ]

    if set(keys) & set(_new_keys):
        remaining = set(keys) & set(_new_keys)
        ret["result"] = False
        ret["comment"].append(
            "State check revealed the following keys could not be deleted: "
            + ", ".join(remaining)
        )
        ret["changes"]["deleted"] = list(
            set(ret["changes"]["deleted"]) - set(_new_keys)
        )

    elif (
        not _new_keys
        and keyring
        and keyring_absent_if_empty
        and __salt__["file.file_exists"](keyring)
    ):
        __salt__["file.remove"](keyring)
        ret["comment"].append(f"Removed empty keyring file {keyring}")
        ret["changes"]["removed"] = keyring

    ret["comment"] = "\n".join(ret["comment"])
    return ret


def _import_data(key, refresh, user, gnupghome, keyring, text=None, path=None):
    has_key = __salt__["gpg.read_key"](
        text=text, path=path, keyid=key, gnupghome=gnupghome, user=user
    )
    if has_key:
        is_expired = has_key[0].get("expired")
        # Ensure we still import the expired key if it's not present
        if not is_expired or not refresh:
            log.debug("Passed text contains key %s", key)
            return __salt__["gpg.import_key"](
                text=text,
                filename=path,
                user=user,
                gnupghome=gnupghome,
                keyring=keyring,
                select=key,
            )
        raise KeyNotContained(f"Passed text contained the key {key}, but it's expired")
    raise KeyNotContained(f"Passed text did not contain the requested key {key}")



from __future__ import annotations
import builtins
import functools
import keyword
import re
import token as T
import tokenize
import unicodedata
import _colorize

from collections import deque
from io import StringIO
from tokenize import TokenInfo as TI
from typing import Iterable, Iterator, Match, NamedTuple, Self

from .types import CharBuffer, CharWidths
from .trace import trace

ANSI_ESCAPE_SEQUENCE = re.compile(r"\x1b\[[ -@]*[A-~]")
ZERO_WIDTH_BRACKET = re.compile(r"\x01.*?\x02")
ZERO_WIDTH_TRANS = str.maketrans({"\x01": "", "\x02": ""})
IDENTIFIERS_AFTER = {"def", "class"}
KEYWORD_CONSTANTS = {"True", "False", "None"}
BUILTINS = {str(name) for name in dir(builtins) if not name.startswith('_')}


def THEME(**kwargs):
    # Not cached: the user can modify the theme inside the interactive session.
    return _colorize.get_theme(**kwargs).syntax


class Span(NamedTuple):
    """Span indexing that's inclusive on both ends."""

    start: int
    end: int

    @classmethod
    def from_re(cls, m: Match[str], group: int | str) -> Self:
        re_span = m.span(group)
        return cls(re_span[0], re_span[1] - 1)

    @classmethod
    def from_token(cls, token: TI, line_len: list[int]) -> Self:
        end_offset = -1
        if (token.type in {T.FSTRING_MIDDLE, T.TSTRING_MIDDLE}
            and token.string.endswith(("{", "}"))):
            # gh-134158: a visible trailing brace comes from a double brace in input
            end_offset += 1

        return cls(
            line_len[token.start[0] - 1] + token.start[1],
            line_len[token.end[0] - 1] + token.end[1] + end_offset,
        )


class ColorSpan(NamedTuple):
    span: Span
    tag: str


@functools.cache
def str_width(c: str) -> int:
    if ord(c) < 128:
        return 1
    # gh-139246 for zero-width joiner and combining characters
    if unicodedata.combining(c):
        return 0
    category = unicodedata.category(c)
    if category == "Cf" and c != "\u00ad":
        return 0
    w = unicodedata.east_asian_width(c)
    if w in ("N", "Na", "H", "A"):
        return 1
    return 2


def wlen(s: str) -> int:
    if len(s) == 1 and s != "\x1a":
        return str_width(s)
    length = sum(str_width(i) for i in s)
    # remove lengths of any escape sequences
    sequence = ANSI_ESCAPE_SEQUENCE.findall(s)
    ctrl_z_cnt = s.count("\x1a")
    return length - sum(len(i) for i in sequence) + ctrl_z_cnt


def unbracket(s: str, including_content: bool = False) -> str:
    r"""Return `s` with \001 and \002 characters removed.

    If `including_content` is True, content between \001 and \002 is also
    stripped.
    """
    if including_content:
        return ZERO_WIDTH_BRACKET.sub("", s)
    return s.translate(ZERO_WIDTH_TRANS)


def gen_colors(buffer: str) -> Iterator[ColorSpan]:
    """Returns a list of index spans to color using the given color tag.

    The input `buffer` should be a valid start of a Python code block, i.e.
    it cannot be a block starting in the middle of a multiline string.
    """
    sio = StringIO(buffer)
    line_lengths = [0] + [len(line) for line in sio.readlines()]
    # make line_lengths cumulative
    for i in range(1, len(line_lengths)):
        line_lengths[i] += line_lengths[i-1]

    sio.seek(0)
    gen = tokenize.generate_tokens(sio.readline)
    last_emitted: ColorSpan | None = None
    try:
        for color in gen_colors_from_token_stream(gen, line_lengths):
            yield color
            last_emitted = color
    except SyntaxError:
        return
    except tokenize.TokenError as te:
        yield from recover_unterminated_string(
            te, line_lengths, last_emitted, buffer
        )


def recover_unterminated_string(
    exc: tokenize.TokenError,
    line_lengths: list[int],
    last_emitted: ColorSpan | None,
    buffer: str,
) -> Iterator[ColorSpan]:
    msg, loc = exc.args
    if loc is None:
        return

    line_no, column = loc

    if msg.startswith(
        (
            "unterminated string literal",
            "unterminated f-string literal",
            "unterminated t-string literal",
            "EOF in multi-line string",
            "unterminated triple-quoted f-string literal",
            "unterminated triple-quoted t-string literal",
        )
    ):
        start = line_lengths[line_no - 1] + column - 1
        end = line_lengths[-1] - 1

        # in case FSTRING_START was already emitted
        if last_emitted and start <= last_emitted.span.start:
            trace("before last emitted = {s}", s=start)
            start = last_emitted.span.end + 1

        span = Span(start, end)
        trace("yielding span {a} -> {b}", a=span.start, b=span.end)
        yield ColorSpan(span, "string")
    else:
        trace(
            "unhandled token error({buffer}) = {te}",
            buffer=repr(buffer),
            te=str(exc),
        )


def gen_colors_from_token_stream(
    token_generator: Iterator[TI],
    line_lengths: list[int],
) -> Iterator[ColorSpan]:
    token_window = prev_next_window(token_generator)

    is_def_name = False
    bracket_level = 0
    for prev_token, token, next_token in token_window:
        assert token is not None
        if token.start == token.end:
            continue

        match token.type:
            case (
                T.STRING
                | T.FSTRING_START | T.FSTRING_MIDDLE | T.FSTRING_END
                | T.TSTRING_START | T.TSTRING_MIDDLE | T.TSTRING_END
            ):
                span = Span.from_token(token, line_lengths)
                yield ColorSpan(span, "string")
            case T.COMMENT:
                span = Span.from_token(token, line_lengths)
                yield ColorSpan(span, "comment")
            case T.NUMBER:
                span = Span.from_token(token, line_lengths)
                yield ColorSpan(span, "number")
            case T.OP:
                if token.string in "([{":
                    bracket_level += 1
                elif token.string in ")]}":
                    bracket_level -= 1
                span = Span.from_token(token, line_lengths)
                yield ColorSpan(span, "op")
            case T.NAME:
                if is_def_name:
                    is_def_name = False
                    span = Span.from_token(token, line_lengths)
                    yield ColorSpan(span, "definition")
                elif keyword.iskeyword(token.string):
                    span_cls = "keyword"
                    if token.string in KEYWORD_CONSTANTS:
                        span_cls = "keyword_constant"
                    span = Span.from_token(token, line_lengths)
                    yield ColorSpan(span, span_cls)
                    if token.string in IDENTIFIERS_AFTER:
                        is_def_name = True
                elif (
                    keyword.issoftkeyword(token.string)
                    and bracket_level == 0
                    and is_soft_keyword_used(prev_token, token, next_token)
                ):
                    span = Span.from_token(token, line_lengths)
                    yield ColorSpan(span, "soft_keyword")
                elif (
                    token.string in BUILTINS
                    and not (prev_token and prev_token.exact_type == T.DOT)
                ):
                    span = Span.from_token(token, line_lengths)
                    yield ColorSpan(span, "builtin")


keyword_first_sets_match = {"False", "None", "True", "await", "lambda", "not"}
keyword_first_sets_case = {"False", "None", "True"}


def is_soft_keyword_used(*tokens: TI | None) -> bool:
    """Returns True if the current token is a keyword in this context.

    For the `*tokens` to match anything, they have to be a three-tuple of
    (previous, current, next).
    """
    trace("is_soft_keyword_used{t}", t=tokens)
    match tokens:
        case (
            None | TI(T.NEWLINE) | TI(T.INDENT) | TI(string=":"),
            TI(string="match"),
            TI(T.NUMBER | T.STRING | T.FSTRING_START | T.TSTRING_START)
            | TI(T.OP, string="(" | "*" | "[" | "{" | "~" | "...")
        ):
            return True
        case (
            None | TI(T.NEWLINE) | TI(T.INDENT) | TI(string=":"),
            TI(string="match"),
            TI(T.NAME, string=s)
        ):
            if keyword.iskeyword(s):
                return s in keyword_first_sets_match
            return True
        case (
            None | TI(T.NEWLINE) | TI(T.INDENT) | TI(T.DEDENT) | TI(string=":"),
            TI(string="case"),
            TI(T.NUMBER | T.STRING | T.FSTRING_START | T.TSTRING_START)
            | TI(T.OP, string="(" | "*" | "-" | "[" | "{")
        ):
            return True
        case (
            None | TI(T.NEWLINE) | TI(T.INDENT) | TI(T.DEDENT) | TI(string=":"),
            TI(string="case"),
            TI(T.NAME, string=s)
        ):
            if keyword.iskeyword(s):
                return s in keyword_first_sets_case
            return True
        case (TI(string="case"), TI(string="_"), TI(string=":")):
            return True
        case (
            None | TI(T.NEWLINE) | TI(T.INDENT) | TI(T.DEDENT) | TI(string=":"),
            TI(string="type"),
            TI(T.NAME, string=s)
        ):
            return not keyword.iskeyword(s)
        case _:
            return False


def disp_str(
    buffer: str,
    colors: list[ColorSpan] | None = None,
    start_index: int = 0,
    force_color: bool = False,
) -> tuple[CharBuffer, CharWidths]:
    r"""Decompose the input buffer into a printable variant with applied colors.

    Returns a tuple of two lists:
    - the first list is the input buffer, character by character, with color
      escape codes added (while those codes contain multiple ASCII characters,
      each code is considered atomic *and is attached for the corresponding
      visible character*);
    - the second list is the visible width of each character in the input
      buffer.

    Note on colors:
    - The `colors` list, if provided, is partially consumed within. We're using
      a list and not a generator since we need to hold onto the current
      unfinished span between calls to disp_str in case of multiline strings.
    - The `colors` list is computed from the start of the input block. `buffer`
      is only a subset of that input block, a single line within. This is why
      we need `start_index` to inform us which position is the start of `buffer`
      actually within user input. This allows us to match color spans correctly.

    Examples:
    >>> utils.disp_str("a = 9")
    (['a', ' ', '=', ' ', '9'], [1, 1, 1, 1, 1])

    >>> line = "while 1:"
    >>> colors = list(utils.gen_colors(line))
    >>> utils.disp_str(line, colors=colors)
    (['\x1b[1;34mw', 'h', 'i', 'l', 'e\x1b[0m', ' ', '1', ':'], [1, 1, 1, 1, 1, 1, 1, 1])

    """
    chars: CharBuffer = []
    char_widths: CharWidths = []

    if not buffer:
        return chars, char_widths

    while colors and colors[0].span.end < start_index:
        # move past irrelevant spans
        colors.pop(0)

    theme = THEME(force_color=force_color)
    pre_color = ""
    post_color = ""
    if colors and colors[0].span.start < start_index:
        # looks like we're continuing a previous color (e.g. a multiline str)
        pre_color = theme[colors[0].tag]

    for i, c in enumerate(buffer, start_index):
        if colors and colors[0].span.start == i:  # new color starts now
            pre_color = theme[colors[0].tag]

        if c == "\x1a":  # CTRL-Z on Windows
            chars.append(c)
            char_widths.append(2)
        elif ord(c) < 128:
            chars.append(c)
            char_widths.append(1)
        elif unicodedata.category(c).startswith("C"):
            c = r"\u%04x" % ord(c)
            chars.append(c)
            char_widths.append(len(c))
        else:
            chars.append(c)
            char_widths.append(str_width(c))

        if colors and colors[0].span.end == i:  # current color ends now
            post_color = theme.reset
            colors.pop(0)

        chars[-1] = pre_color + chars[-1] + post_color
        pre_color = ""
        post_color = ""

    if colors and colors[0].span.start < i and colors[0].span.end > i:
        # even though the current color should be continued, reset it for now.
        # the next call to `disp_str()` will revive it.
        chars[-1] += theme.reset

    return chars, char_widths


def prev_next_window[T](
    iterable: Iterable[T]
) -> Iterator[tuple[T | None, ...]]:
    """Generates three-tuples of (previous, current, next) items.

    On the first iteration previous is None. On the last iteration next
    is None. In case of exception next is None and the exception is re-raised
    on a subsequent next() call.

    Inspired by `sliding_window` from `itertools` recipes.
    """

    iterator = iter(iterable)
    window = deque((None, next(iterator)), maxlen=3)
    try:
        for x in iterator:
            window.append(x)
            yield tuple(window)
    except Exception:
        raise
    finally:
        window.append(None)
        yield tuple(window)



r"""
Manage the Windows registry
===========================

Many python developers think of registry keys as if they were python keys in a
dictionary which is not the case. The windows registry is broken down into the
following components:

Hives
-----

This is the top level of the registry. They all begin with HKEY.

    - HKEY_CLASSES_ROOT (HKCR)
    - HKEY_CURRENT_USER (HKCU)
    - HKEY_LOCAL MACHINE (HKLM)
    - HKEY_USER (HKU)
    - HKEY_CURRENT_CONFIG

Keys
----

Hives contain keys. These are basically the folders beneath the hives. They can
contain any number of subkeys.

When passing the hive\key values they must be quoted correctly depending on the
backslashes being used (``\`` vs ``\\``). The way backslashes are handled in
the state file is different from the way they are handled when working on the
CLI. The following are valid methods of passing the hive\key:

Using single backslashes:
    HKLM\SOFTWARE\Python
    'HKLM\SOFTWARE\Python'

Using double backslashes:
    "HKLM\\SOFTWARE\\Python"

Values or Entries
-----------------

Values or Entries are the name/data pairs beneath the keys and subkeys. All keys
have a default name/data pair. The name is ``(Default)`` with a displayed value
of ``(value not set)``. The actual value is Null.

Example
-------

The following example is taken from the windows startup portion of the registry:

.. code-block:: text

    [HKEY_LOCAL_MACHINE\SOFTWARE\Microsoft\Windows\CurrentVersion\Run]
    "RTHDVCPL"="\"C:\\Program Files\\Realtek\\Audio\\HDA\\RtkNGUI64.exe\" -s"
    "NvBackend"="\"C:\\Program Files (x86)\\NVIDIA Corporation\\Update Core\\NvBackend.exe\""
    "BTMTrayAgent"="rundll32.exe \"C:\\Program Files (x86)\\Intel\\Bluetooth\\btmshellex.dll\",TrayApp"

In this example these are the values for each:

Hive:
    ``HKEY_LOCAL_MACHINE``

Key and subkeys:
    ``SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Run``

Value:

- There are 3 value names: ``RTHDVCPL``, ``NvBackend``, and ``BTMTrayAgent``
- Each value name has a corresponding value

"""

import logging

import salt.utils.stringutils
import salt.utils.win_dacl as dacl
import salt.utils.win_reg as reg

log = logging.getLogger(__name__)


def __virtual__():
    """
    Load this state if the reg module exists
    """
    return "reg"


def _parse_key(key):
    """
    split the hive from the key
    """
    splt = key.split("\\")
    hive = splt.pop(0)
    key = "\\".join(splt)
    return hive, key


def present(
    name,
    vname=None,
    vdata=None,
    vtype="REG_SZ",
    use_32bit_registry=False,
    win_owner=None,
    win_perms=None,
    win_deny_perms=None,
    win_inheritance=True,
    win_perms_reset=False,
):
    r"""
    Ensure a registry key or value is present.

    Args:

        name (str):
            A string value representing the full path of the key to include the
            HIVE, Key, and all Subkeys. For example:

            ``HKEY_LOCAL_MACHINE\SOFTWARE\Salt``

            Valid hive values include:

                - HKEY_CURRENT_USER or HKCU
                - HKEY_LOCAL_MACHINE or HKLM
                - HKEY_USERS or HKU

        vname (str):
            The name of the value you'd like to create beneath the Key. If this
            parameter is not passed it will assume you want to set the
            ``(Default)`` value

        vdata (str, int, list, bytes):
            The value you'd like to set. If a value name (``vname``) is passed,
            this will be the data for that value name. If not, this will be the
            ``(Default)`` value for the key.

            The type of data this parameter expects is determined by the value
            type specified in ``vtype``. The correspondence is as follows:

                - REG_BINARY: Binary data (str in Py2, bytes in Py3)
                - REG_DWORD: int
                - REG_EXPAND_SZ: str
                - REG_MULTI_SZ: list of str
                - REG_QWORD: int
                - REG_SZ: str

                .. note::
                    When setting REG_BINARY, string data will be converted to
                    binary automatically. To pass binary data, use the built-in
                    yaml tag ``!!binary`` to denote the actual binary
                    characters. For example, the following lines will both set
                    the same data in the registry:

                    - ``vdata: Salty Test``
                    - ``vdata: !!binary U2FsdHkgVGVzdA==\n``

                    For more information about the ``!!binary`` tag see
                    `here <http://yaml.org/type/binary.html>`_

            .. note::
                The type for the ``(Default)`` value is always REG_SZ and cannot
                be changed. This parameter is optional. If not passed, the Key
                will be created with no associated item/value pairs.

        vtype (str):
            The value type for the data you wish to store in the registry. Valid
            values are:

                - REG_BINARY
                - REG_DWORD
                - REG_EXPAND_SZ
                - REG_MULTI_SZ
                - REG_QWORD
                - REG_SZ (Default)

        use_32bit_registry (bool):
            Use the 32bit portion of the registry. Applies only to 64bit
            windows. 32bit Windows will ignore this parameter. Default is False.

        win_owner (str):
            The owner of the registry key. If this is not passed, the account
            under which Salt is running will be used.

            .. note::
                Owner is set for the key that contains the value/data pair. You
                cannot set ownership on value/data pairs themselves.

            .. versionadded:: 2019.2.0

        win_perms (dict):
            A dictionary containing permissions to grant and their propagation.
            If not passed the 'Grant` permissions will not be modified.

            .. note::
                Permissions are set for the key that contains the value/data
                pair. You cannot set permissions on value/data pairs themselves.

            For each user specify the account name, with a sub dict for the
            permissions to grant and the 'Applies to' setting. For example:
            ``{'Administrators': {'perms': 'full_control', 'applies_to':
            'this_key_subkeys'}}``. ``perms`` must be specified.

            Registry permissions are specified using the ``perms`` key. You can
            specify a single basic permission or a list of advanced perms. The
            following are valid perms:

                Basic (passed as a string):
                    - full_control
                    - read
                    - write

                Advanced (passed as a list):
                    - delete
                    - query_value
                    - set_value
                    - create_subkey
                    - enum_subkeys
                    - notify
                    - create_link
                    - read_control
                    - write_dac
                    - write_owner

            The 'Applies to' setting is optional. It is specified using the
            ``applies_to`` key. If not specified ``this_key_subkeys`` is used.
            Valid options are:

                Applies to settings:
                    - this_key_only
                    - this_key_subkeys
                    - subkeys_only

            .. versionadded:: 2019.2.0

        win_deny_perms (dict):
            A dictionary containing permissions to deny and their propagation.
            If not passed the `Deny` permissions will not be modified.

            .. note::
                Permissions are set for the key that contains the value/data
                pair. You cannot set permissions on value/data pairs themselves.

            Valid options are the same as those specified in ``win_perms``

            .. note::
                'Deny' permissions always take precedence over 'grant'
                 permissions.

            .. versionadded:: 2019.2.0

        win_inheritance (bool):
            ``True`` to inherit permissions from the parent key. ``False`` to
            disable inheritance. Default is ``True``.

            .. note::
                Inheritance is set for the key that contains the value/data
                pair. You cannot set inheritance on value/data pairs themselves.

            .. versionadded:: 2019.2.0

        win_perms_reset (bool):
            If ``True`` the existing DACL will be cleared and replaced with the
            settings defined in this function. If ``False``, new entries will be
            appended to the existing DACL. Default is ``False``

            .. note::
                Perms are reset for the key that contains the value/data pair.
                You cannot set permissions on value/data pairs themselves.

            .. versionadded:: 2019.2.0

    Returns:
        dict: A dictionary showing the results of the registry operation.

    Example:

    The following example will set the ``(Default)`` value for the
    ``SOFTWARE\Salt`` key in the ``HKEY_CURRENT_USER`` hive to
    ``2016.3.1``:

    .. code-block:: yaml

        HKEY_CURRENT_USER\SOFTWARE\Salt:
          reg.present:
            - vdata: 2016.3.1

    Example:

    The following example will set the value for the ``version`` entry under
    the ``SOFTWARE\Salt`` key in the ``HKEY_CURRENT_USER`` hive to
    ``2016.3.1``. The value will be reflected in ``Wow6432Node``:

    .. code-block:: yaml

        HKEY_CURRENT_USER\SOFTWARE\Salt:
          reg.present:
            - vname: version
            - vdata: 2016.3.1

    In the above example the path is interpreted as follows:

        - ``HKEY_CURRENT_USER`` is the hive
        - ``SOFTWARE\Salt`` is the key
        - ``vname`` is the value name ('version') that will be created under the key
        - ``vdata`` is the data that will be assigned to 'version'

    Example:

    Binary data can be set in two ways. The following two examples will set
    a binary value of ``Salty Test``

    .. code-block:: yaml

        no_conversion:
          reg.present:
            - name: HKLM\SOFTWARE\SaltTesting
            - vname: test_reg_binary_state
            - vdata: Salty Test
            - vtype: REG_BINARY

        conversion:
          reg.present:
            - name: HKLM\SOFTWARE\SaltTesting
            - vname: test_reg_binary_state_with_tag
            - vdata: !!binary U2FsdHkgVGVzdA==\n
            - vtype: REG_BINARY

    Example:

    To set a ``REG_MULTI_SZ`` value:

    .. code-block:: yaml

        reg_multi_sz:
          reg.present:
            - name: HKLM\SOFTWARE\Salt
            - vname: reg_multi_sz
            - vdata:
              - list item 1
              - list item 2

    Example:

    To ensure a key is present and has permissions:

    .. code-block:: yaml

        set_key_permissions:
          reg.present:
            - name: HKLM\SOFTWARE\Salt
            - vname: version
            - vdata: 2016.3.1
            - win_owner: Administrators
            - win_perms:
                jsnuffy:
                  perms: full_control
                sjones:
                  perms:
                    - read_control
                    - enum_subkeys
                    - query_value
                  applies_to:
                    - this_key_only
            - win_deny_perms:
                bsimpson:
                  perms: full_control
                  applies_to: this_key_subkeys
            - win_inheritance: True
            - win_perms_reset: True
    """
    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    hive, key = _parse_key(name)

    # Determine what to do
    reg_current = reg.read_value(
        hive=hive, key=key, vname=vname, use_32bit_registry=use_32bit_registry
    )

    # Cast the vdata according to the vtype
    vdata_decoded = reg.cast_vdata(vdata=vdata, vtype=vtype)

    # Check if the key already exists
    # If so, check perms
    # We check `vdata` and `success` because `vdata` can be None
    if vdata_decoded == reg_current["vdata"] and reg_current["success"]:
        ret["comment"] = "{} in {} is already present".format(
            salt.utils.stringutils.to_unicode(vname, "utf-8") if vname else "(Default)",
            salt.utils.stringutils.to_unicode(name, "utf-8"),
        )
        return dacl.check_perms(
            obj_name="\\".join([hive, key]),
            obj_type="registry32" if use_32bit_registry else "registry",
            ret=ret,
            owner=win_owner,
            grant_perms=win_perms,
            deny_perms=win_deny_perms,
            inheritance=win_inheritance,
            reset=win_perms_reset,
        )

    add_change = {
        "Key": rf"{hive}\{key}",
        "Entry": "{}".format(
            salt.utils.stringutils.to_unicode(vname, "utf-8") if vname else "(Default)"
        ),
        "Value": vdata_decoded,
        "Owner": win_owner,
        "Perms": {"Grant": win_perms, "Deny": win_deny_perms},
        "Inheritance": win_inheritance,
    }

    # Check for test option
    if __opts__["test"]:
        ret["result"] = None
        ret["changes"] = {"reg": {"Will add": add_change}}
        ret["comment"] = rf"Will add {vname} to {hive}\{key}"
        return ret

    # Configure the value
    ret["result"] = reg.set_value(
        hive=hive,
        key=key,
        vname=vname,
        vdata=vdata,
        vtype=vtype,
        use_32bit_registry=use_32bit_registry,
    )

    if not ret["result"]:
        ret["changes"] = {}
        ret["comment"] = rf"Failed to add {vname} to {hive}\{key}"
    else:
        ret["changes"] = {"reg": {"Added": add_change}}
        ret["comment"] = rf"Added {vname} to {hive}\{key}"

    if ret["result"]:
        ret = dacl.check_perms(
            obj_name="\\".join([hive, key]),
            obj_type="registry32" if use_32bit_registry else "registry",
            ret=ret,
            owner=win_owner,
            grant_perms=win_perms,
            deny_perms=win_deny_perms,
            inheritance=win_inheritance,
            reset=win_perms_reset,
        )

    return ret


def absent(name, vname=None, use_32bit_registry=False):
    r"""
    Ensure a registry value is removed. To remove a key use key_absent.

    Args:
        name (str):
            A string value representing the full path of the key to include the
            HIVE, Key, and all Subkeys. For example:

            ``HKEY_LOCAL_MACHINE\SOFTWARE\Salt``

            Valid hive values include:

                - HKEY_CURRENT_USER or HKCU
                - HKEY_LOCAL_MACHINE or HKLM
                - HKEY_USERS or HKU

        vname (str):
            The name of the value you'd like to create beneath the Key. If this
            parameter is not passed it will assume you want to set the
            ``(Default)`` value

        use_32bit_registry (bool):
            Use the 32bit portion of the registry. Applies only to 64bit
            windows. 32bit Windows will ignore this parameter. Default is False.

    Returns:
        dict: A dictionary showing the results of the registry operation.

    CLI Example:

        .. code-block:: yaml

            'HKEY_CURRENT_USER\SOFTWARE\Salt':
              reg.absent
                - vname: version

        In the above example the value named ``version`` will be removed from
        the SOFTWARE\\Salt key in the HKEY_CURRENT_USER hive. If ``vname`` was
        not passed, the ``(Default)`` value would be deleted.
    """
    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    hive, key = _parse_key(name)

    # Determine what to do
    reg_check = reg.read_value(
        hive=hive, key=key, vname=vname, use_32bit_registry=use_32bit_registry
    )
    if not reg_check["success"] or reg_check["vdata"] == "(value not set)":
        ret["comment"] = f"{name} is already absent"
        return ret

    remove_change = {
        "Key": rf"{hive}\{key}",
        "Entry": "{}".format(vname if vname else "(Default)"),
    }

    # Check for test option
    if __opts__["test"]:
        ret["result"] = None
        ret["changes"] = {"reg": {"Will remove": remove_change}}
        ret["comment"] = rf"Will remove {vname} to {hive}\{key}"
        return ret

    # Delete the value
    ret["result"] = reg.delete_value(
        hive=hive, key=key, vname=vname, use_32bit_registry=use_32bit_registry
    )
    if not ret["result"]:
        ret["changes"] = {}
        ret["comment"] = rf"Failed to remove {key} from {hive}"
    else:
        ret["changes"] = {"reg": {"Removed": remove_change}}
        ret["comment"] = rf"Removed {key} from {hive}"

    return ret


def key_absent(name, use_32bit_registry=False):
    r"""
    .. versionadded:: 2015.5.4

    Ensure a registry key is removed. This will remove the key, subkeys, and all
    value entries.

    Args:

        name (str):
            A string representing the full path to the key to be removed to
            include the hive and the keypath. The hive can be any of the
            following:

                - HKEY_LOCAL_MACHINE or HKLM
                - HKEY_CURRENT_USER or HKCU
                - HKEY_USER or HKU

        use_32bit_registry (bool):
            Use the 32bit portion of the registry. Applies only to 64bit
            windows. 32bit Windows will ignore this parameter. Default is False.

    Returns:
        dict: A dictionary showing the results of the registry operation.

    CLI Example:

        The following example will delete the ``SOFTWARE\DeleteMe`` key in the
        ``HKEY_LOCAL_MACHINE`` hive including all its subkeys and value pairs.

        .. code-block:: yaml

            remove_key_demo:
              reg.key_absent:
                - name: HKEY_CURRENT_USER\SOFTWARE\DeleteMe

        In the above example the path is interpreted as follows:

            - ``HKEY_CURRENT_USER`` is the hive
            - ``SOFTWARE\DeleteMe`` is the key
    """
    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    hive, key = _parse_key(name)

    # Determine what to do
    if not reg.read_value(hive=hive, key=key, use_32bit_registry=use_32bit_registry)[
        "success"
    ]:
        ret["comment"] = f"{name} is already absent"
        return ret

    ret["changes"] = {"reg": {"Removed": {"Key": rf"{hive}\{key}"}}}
    ret["comment"] = rf"Removed {hive}\{key}"

    # Check for test option
    if __opts__["test"]:
        ret["result"] = None
        ret["changes"] = {"reg": {"Will remove": {"Key": rf"{hive}\{key}"}}}
        ret["comment"] = rf"Will remove {hive}\{key}"
        return ret

    # Delete the value
    reg.delete_key_recursive(hive=hive, key=key, use_32bit_registry=use_32bit_registry)
    if reg.read_value(hive=hive, key=key, use_32bit_registry=use_32bit_registry)[
        "success"
    ]:
        ret["result"] = False
        ret["changes"] = {}
        ret["comment"] = f"Failed to remove registry key {name}"

    return ret



"""
Runner to manage Windows software repo
"""

# WARNING: Any modules imported here must also be added to
# salt/modules/win_repo.py


import logging
import os

import salt.loader
import salt.minion
import salt.template
import salt.utils.files
import salt.utils.gitfs
import salt.utils.msgpack
import salt.utils.path
from salt.exceptions import CommandExecutionError, SaltRenderError

log = logging.getLogger(__name__)

# Global parameters which can be overridden on a per-remote basis
PER_REMOTE_OVERRIDES = ("ssl_verify", "refspecs", "fallback", "proxy")

# Fall back to default per-remote-only. This isn't technically needed since
# salt.utils.gitfs.GitBase.__init__ will default to
# salt.utils.gitfs.PER_REMOTE_ONLY for this value, so this is mainly for
# runners and other modules that import salt.runners.winrepo.
PER_REMOTE_ONLY = salt.utils.gitfs.PER_REMOTE_ONLY
GLOBAL_ONLY = ("branch",)


def _legacy_git():
    return not any(
        (salt.utils.gitfs.GITPYTHON_VERSION, salt.utils.gitfs.PYGIT2_VERSION)
    )


def genrepo(opts=None, fire_event=True):
    """
    Generate winrepo_cachefile based on sls files in the winrepo_dir

    opts
        Specify an alternate opts dict. Should not be used unless this function
        is imported into an execution module.

    fire_event : True
        Fire an event on failure. Only supported on the master.

    CLI Example:

    .. code-block:: bash

        salt-run winrepo.genrepo
    """
    if opts is None:
        opts = __opts__

    winrepo_dir = opts["winrepo_dir"]
    winrepo_cachefile = opts["winrepo_cachefile"]

    ret = {}
    if not os.path.exists(winrepo_dir):
        os.makedirs(winrepo_dir)
    renderers = salt.loader.render(opts, __salt__)
    for root, _, files in salt.utils.path.os_walk(winrepo_dir):
        for name in files:
            if name.endswith(".sls"):
                try:
                    config = salt.template.compile_template(
                        os.path.join(root, name),
                        renderers,
                        opts["renderer"],
                        opts["renderer_blacklist"],
                        opts["renderer_whitelist"],
                    )
                except SaltRenderError as exc:
                    log.debug("Failed to render %s.", os.path.join(root, name))
                    log.debug("Error: %s.", exc)
                    continue
                if config:
                    revmap = {}
                    for pkgname, versions in config.items():
                        log.debug("Compiling winrepo data for package '%s'", pkgname)
                        for version, repodata in versions.items():
                            log.debug(
                                "Compiling winrepo data for %s version %s",
                                pkgname,
                                version,
                            )
                            if not isinstance(version, str):
                                config[pkgname][str(version)] = config[pkgname].pop(
                                    version
                                )
                            if not isinstance(repodata, dict):
                                msg = "Failed to compile {}.".format(
                                    os.path.join(root, name)
                                )
                                log.debug(msg)
                                if fire_event:
                                    try:
                                        __jid_event__.fire_event(
                                            {"error": msg}, "progress"
                                        )
                                    except NameError:
                                        log.error(
                                            "Attempted to fire the an event "
                                            "with the following error, but "
                                            "event firing is not supported: %s",
                                            msg,
                                        )
                                continue
                            revmap[repodata["full_name"]] = pkgname
                    ret.setdefault("repo", {}).update(config)
                    ret.setdefault("name_map", {}).update(revmap)
    with salt.utils.files.fopen(
        os.path.join(winrepo_dir, winrepo_cachefile), "w+b"
    ) as repo:
        repo.write(salt.utils.msgpack.dumps(ret))
    return ret


def update_git_repos(opts=None, clean=False, masterless=False):
    """
    Checkout git repos containing Windows Software Package Definitions

    opts
        Specify an alternate opts dict. Should not be used unless this function
        is imported into an execution module.

    clean : False
        Clean repo cachedirs which are not configured under
        :conf_master:`winrepo_remotes`.

        .. warning::
            This argument should not be set to ``True`` if a mix of git and
            non-git repo definitions are being used, as it will result in the
            non-git repo definitions being removed.

        .. versionadded:: 2015.8.0

    CLI Examples:

    .. code-block:: bash

        salt-run winrepo.update_git_repos
        salt-run winrepo.update_git_repos clean=True
    """
    if opts is None:
        opts = __opts__

    winrepo_dir = opts["winrepo_dir"]
    winrepo_remotes = opts["winrepo_remotes"]

    winrepo_cfg = [
        (winrepo_remotes, winrepo_dir),
        (opts["winrepo_remotes_ng"], opts["winrepo_dir_ng"]),
    ]

    ret = {}
    for remotes, base_dir in winrepo_cfg:
        if _legacy_git():
            # Use legacy code
            winrepo_result = {}
            for remote_info in remotes:
                if "/" in remote_info:
                    targetname = remote_info.split("/")[-1]
                else:
                    targetname = remote_info
                rev = "HEAD"
                # If a revision is specified, use it.
                try:
                    rev, remote_url = remote_info.strip().split()
                except ValueError:
                    remote_url = remote_info
                targetname = targetname.replace(".git", "")
                # GitFS using pygit2 and gitpython place the repo in a
                # subdirectory named `_`. We need to stay consistent when using
                # the legacy method as well
                gittarget = os.path.join(base_dir, targetname, "_")
                if masterless:
                    result = __salt__["state.single"](
                        "git.latest",
                        name=remote_url,
                        rev=rev,
                        branch="winrepo",
                        target=gittarget,
                        force_checkout=True,
                        force_reset=True,
                    )
                    if isinstance(result, list):
                        # Errors were detected
                        raise CommandExecutionError(
                            "Failed up update winrepo remotes: {}".format(
                                "\n".join(result)
                            )
                        )
                    if "name" not in result:
                        # Highstate output dict, the results are actually nested
                        # one level down.
                        key = next(iter(result))
                        result = result[key]
                else:
                    mminion = salt.minion.MasterMinion(opts)
                    result = mminion.functions["state.single"](
                        "git.latest",
                        name=remote_url,
                        rev=rev,
                        branch="winrepo",
                        target=gittarget,
                        force_checkout=True,
                        force_reset=True,
                    )
                    if isinstance(result, list):
                        # Errors were detected
                        raise CommandExecutionError(
                            "Failed to update winrepo remotes: {}".format(
                                "\n".join(result)
                            )
                        )
                    if "name" not in result:
                        # Highstate output dict, the results are actually nested
                        # one level down.
                        key = next(iter(result))
                        result = result[key]
                winrepo_result[result["name"]] = result["result"]
            ret.update(winrepo_result)
        else:
            # New winrepo code utilizing salt.utils.gitfs
            try:
                winrepo = salt.utils.gitfs.WinRepo(
                    opts,
                    remotes,
                    per_remote_overrides=PER_REMOTE_OVERRIDES,
                    per_remote_only=PER_REMOTE_ONLY,
                    global_only=GLOBAL_ONLY,
                    cache_root=base_dir,
                )
                winrepo.fetch_remotes()
                # Since we're not running update(), we need to manually call
                # clear_old_remotes() to remove directories from remotes that
                # have been removed from configuration.
                if clean:
                    winrepo.clear_old_remotes()
                winrepo.checkout()
            except Exception as exc:  # pylint: disable=broad-except
                msg = f"Failed to update winrepo_remotes: {exc}"
                log.error(msg, exc_info_on_loglevel=logging.DEBUG)
                return msg
            ret.update(winrepo.winrepo_dirs)
    return ret



import pathlib
import sys

import salt.utils.files


def bundled():
    """
    Gather run-time information to indicate if we are running from relenv onedir
    """
    if hasattr(sys, "RELENV"):
        return True
    else:
        return False


def pkg_type():
    """
    Utility to find out how Salt was installed.
    """
    # Relenv / onedir installs must win over a sibling ``_pkg.txt`` from a
    # source checkout (which often says ``pip`` when the tree is bind-mounted
    # next to a relenv interpreter).
    if bundled():
        return "onedir"
    pkg_file = pathlib.Path(__file__).parent.parent / "_pkg.txt"
    if pkg_file.is_file():
        with salt.utils.files.fopen(pkg_file) as _fp:
            content = _fp.read()
            if content:
                return content.strip()
    return None



import unittest

from Cryptodome.SelfTest.st_common import list_test_cases
from Cryptodome.SelfTest.loader import load_test_vectors_wycheproof

from Cryptodome.Cipher import AES


class KW_Tests(unittest.TestCase):

    # From RFC3394
    tvs = [
        ("000102030405060708090A0B0C0D0E0F",
         "00112233445566778899AABBCCDDEEFF",
         "1FA68B0A8112B447AEF34BD8FB5A7B829D3E862371D2CFE5"),
        ("000102030405060708090A0B0C0D0E0F1011121314151617",
         "00112233445566778899AABBCCDDEEFF",
         "96778B25AE6CA435F92B5B97C050AED2468AB8A17AD84E5D"),
        ("000102030405060708090A0B0C0D0E0F101112131415161718191A1B1C1D1E1F",
         "00112233445566778899AABBCCDDEEFF",
         "64E8C3F9CE0F5BA263E9777905818A2A93C8191E7D6E8AE7"),
        ("000102030405060708090A0B0C0D0E0F1011121314151617",
         "00112233445566778899AABBCCDDEEFF0001020304050607",
         "031D33264E15D33268F24EC260743EDCE1C6C7DDEE725A936BA814915C6762D2"),
        ("000102030405060708090A0B0C0D0E0F101112131415161718191A1B1C1D1E1F",
         "00112233445566778899AABBCCDDEEFF0001020304050607",
         "A8F9BC1612C68B3FF6E6F4FBE30E71E4769C8B80A32CB8958CD5D17D6B254DA1"),
        ("000102030405060708090A0B0C0D0E0F101112131415161718191A1B1C1D1E1F",
         "00112233445566778899AABBCCDDEEFF000102030405060708090A0B0C0D0E0F",
         "28C9F404C4B810F4CBCCB35CFB87F8263F5786E2D80ED326CBC7F0E71A99F43BFB988B9B7A02DD21"),
    ]

    def test_rfc3394(self):
        for tv in self.tvs:
            kek, pt, ct = [bytes.fromhex(x) for x in tv]

            cipher = AES.new(kek, AES.MODE_KW)
            ct2 = cipher.seal(pt)

            self.assertEqual(ct, ct2)

            cipher = AES.new(kek, AES.MODE_KW)
            pt2 = cipher.unseal(ct)
            self.assertEqual(pt, pt2)

    def test_neg1(self):

        cipher = AES.new(b'-' * 16, AES.MODE_KW)

        with self.assertRaises(ValueError):
            cipher.seal(b'')

        with self.assertRaises(ValueError):
            cipher.seal(b'8' * 17)

    def test_neg2(self):

        cipher = AES.new(b'-' * 16, AES.MODE_KW)
        ct = bytearray(cipher.seal(b'7' * 16))

        cipher = AES.new(b'-' * 16, AES.MODE_KW)
        cipher.unseal(ct)

        cipher = AES.new(b'-' * 16, AES.MODE_KW)
        ct[0] ^= 0xFF
        with self.assertRaises(ValueError):
            cipher.unseal(ct)


class KW_Wycheproof(unittest.TestCase):

    def setUp(self):
        self.vectors = load_test_vectors_wycheproof(("Cipher", "wycheproof"),
                                                    "kw_test.json",
                                                    "Wycheproof tests for KW")

    def test_wycheproof(self):

        if not self.vectors:
            self.skipTest("No test vectors available")

        for vector in self.vectors:
            with self.subTest(testId=vector.id):
                cipher = AES.new(vector.key, AES.MODE_KW)

                try:
                    cipher.seal(vector.msg)
                except ValueError:
                    if vector.valid:
                        raise
                    continue

                cipher = AES.new(vector.key, AES.MODE_KW)
                try:
                    pt = cipher.unseal(vector.ct)
                except ValueError:
                    if vector.valid:
                        raise
                    continue

                self.assertEqual(pt, vector.msg)


class KWP_Tests(unittest.TestCase):

    tvs = [
        ("5840df6e29b02af1ab493b705bf16ea1ae8338f4dcc176a8",
         "c37b7e6492584340bed12207808941155068f738",
         "138bdeaa9b8fa7fc61f97742e72248ee5ae6ae5360d1ae6a5f54f373fa543b6a"),
        ("5840df6e29b02af1ab493b705bf16ea1ae8338f4dcc176a8",
         "466f7250617369",
         "afbeb0f07dfbf5419200f2ccb50bb24f"),
    ]

    def test_rfc5649(self):
        for tv in self.tvs:
            kek, pt, ct = [bytes.fromhex(x) for x in tv]

            cipher = AES.new(kek, AES.MODE_KWP)
            ct2 = cipher.seal(pt)

            self.assertEqual(ct, ct2)

            cipher = AES.new(kek, AES.MODE_KWP)
            pt2 = cipher.unseal(ct)
            self.assertEqual(pt, pt2)


class KWP_Wycheproof(unittest.TestCase):

    def setUp(self):
        self.vectors = load_test_vectors_wycheproof(("Cipher", "wycheproof"),
                                                    "kwp_test.json",
                                                    "Wycheproof tests for KWP")

    def test_wycheproof(self):

        if not self.vectors:
            self.skipTest("No test vectors available")

        for vector in self.vectors:
            with self.subTest(testId=vector.id):
                cipher = AES.new(vector.key, AES.MODE_KWP)

                try:
                    cipher.seal(vector.msg)
                except ValueError:
                    if vector.valid and not vector.warning:
                        raise
                    continue

                cipher = AES.new(vector.key, AES.MODE_KWP)
                try:
                    pt = cipher.unseal(vector.ct)
                except ValueError:
                    if vector.valid and not vector.warning:
                        raise
                    continue

                self.assertEqual(pt, vector.msg)


def get_tests(config={}):
    tests = []
    tests += list_test_cases(KW_Tests)
    tests += list_test_cases(KWP_Tests)
    tests += list_test_cases(KW_Wycheproof)
    tests += list_test_cases(KWP_Wycheproof)
    return tests


if __name__ == '__main__':
    def suite():
        return unittest.TestSuite(get_tests())
    unittest.main(defaultTest='suite')



"""

This module allows you to control the power settings of a windows minion via
powercfg.

.. versionadded:: 2015.8.0

.. code-block:: yaml

    # Set timeout to 30 minutes on battery power
    monitor:
        powercfg.set_timeout:
            - value: 30
            - power: dc
"""

import logging

import salt.utils.data
import salt.utils.platform

log = logging.getLogger(__name__)

__virtualname__ = "powercfg"


def __virtual__():
    """
    Only work on Windows
    """
    if not salt.utils.platform.is_windows():
        return False, "PowerCFG: Module only works on Windows"
    return __virtualname__


def set_timeout(name, value, power="ac", scheme=None):
    """
    Set the sleep timeouts of specific items such as disk, monitor, etc.

    Args:

        name (str)
            The setting to change, can be one of the following:

                - ``monitor``
                - ``disk``
                - ``standby``
                - ``hibernate``

        value (int):
            The amount of time in minutes before the item will timeout

        power (str):
            Set the value for AC or DC power. Default is ``ac``. Valid options
            are:

                - ``ac`` (AC Power)
                - ``dc`` (Battery)

        scheme (str):
            The scheme to use, leave as ``None`` to use the current. Default is
            ``None``. This can be the GUID or the Alias for the Scheme. Known
            Aliases are:

                - ``SCHEME_BALANCED`` - Balanced
                - ``SCHEME_MAX`` - Power saver
                - ``SCHEME_MIN`` - High performance

    CLI Example:

    .. code-block:: yaml

        # Set monitor timeout to 30 minutes on Battery
        monitor:
          powercfg.set_timeout:
            - value: 30
            - power: dc

        # Set disk timeout to 10 minutes on AC Power
        disk:
          powercfg.set_timeout:
            - value: 10
            - power: ac
    """
    ret = {"name": name, "result": True, "comment": "", "changes": {}}

    # Validate name values
    name = name.lower()
    if name not in ["monitor", "disk", "standby", "hibernate"]:
        ret["result"] = False
        ret["comment"] = f'"{name}" is not a valid setting'
        log.debug(ret["comment"])
        return ret

    # Validate power values
    power = power.lower()
    if power not in ["ac", "dc"]:
        ret["result"] = False
        ret["comment"] = f'"{power}" is not a power type'
        log.debug(ret["comment"])
        return ret

    # Get current settings
    old = __salt__[f"powercfg.get_{name}_timeout"](scheme=scheme)

    # Check current settings
    if old[power] == value:
        ret["comment"] = "{} timeout on {} power is already set to {}".format(
            name.capitalize(), power.upper(), value
        )
        return ret
    else:
        ret["comment"] = "{} timeout on {} power will be set to {}".format(
            name.capitalize(), power.upper(), value
        )

    # Check for test=True
    if __opts__["test"]:
        ret["result"] = None
        return ret

    # Set the timeout value
    __salt__[f"powercfg.set_{name}_timeout"](timeout=value, power=power, scheme=scheme)

    # Get the setting after the change
    new = __salt__[f"powercfg.get_{name}_timeout"](scheme=scheme)

    changes = salt.utils.data.compare_dicts(old, new)

    if changes:
        ret["changes"] = {name: changes}
        ret["comment"] = "{} timeout on {} power set to {}".format(
            name.capitalize(), power.upper(), value
        )
        log.debug(ret["comment"])
    else:
        ret["changes"] = {}
        ret["comment"] = "Failed to set {} timeout on {} power to {}".format(
            name, power.upper(), value
        )
        log.debug(ret["comment"])
        ret["result"] = False

    return ret



"""Module with a simple buffer implementation using the memory manager"""
import sys

__all__ = ["SlidingWindowMapBuffer"]


class SlidingWindowMapBuffer:

    """A buffer like object which allows direct byte-wise object and slicing into
    memory of a mapped file. The mapping is controlled by the provided cursor.

    The buffer is relative, that is if you map an offset, index 0 will map to the
    first byte at the offset you used during initialization or begin_access

    **Note:** Although this type effectively hides the fact that there are mapped windows
    underneath, it can unfortunately not be used in any non-pure python method which
    needs a buffer or string"""
    __slots__ = (
        '_c',           # our cursor
        '_size',        # our supposed size
    )

    def __init__(self, cursor=None, offset=0, size=sys.maxsize, flags=0):
        """Initialize the instance to operate on the given cursor.
        :param cursor: if not None, the associated cursor to the file you want to access
            If None, you have call begin_access before using the buffer and provide a cursor
        :param offset: absolute offset in bytes
        :param size: the total size of the mapping. Defaults to the maximum possible size
            From that point on, the __len__ of the buffer will be the given size or the file size.
            If the size is larger than the mappable area, you can only access the actually available
            area, although the length of the buffer is reported to be your given size.
            Hence it is in your own interest to provide a proper size !
        :param flags: Additional flags to be passed to os.open
        :raise ValueError: if the buffer could not achieve a valid state"""
        self._c = cursor
        if cursor and not self.begin_access(cursor, offset, size, flags):
            raise ValueError("Failed to allocate the buffer - probably the given offset is out of bounds")
        # END handle offset

    def __del__(self):
        self.end_access()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.end_access()

    def __len__(self):
        return self._size

    def __getitem__(self, i):
        if isinstance(i, slice):
            return self.__getslice__(i.start or 0, i.stop or self._size)
        c = self._c
        assert c.is_valid()
        if i < 0:
            i = self._size + i
        if not c.includes_ofs(i):
            c.use_region(i, 1)
        # END handle region usage
        return c.buffer()[i - c.ofs_begin()]

    def __getslice__(self, i, j):
        c = self._c
        # fast path, slice fully included - safes a concatenate operation and
        # should be the default
        assert c.is_valid()
        if i < 0:
            i = self._size + i
        if j == sys.maxsize:
            j = self._size
        if j < 0:
            j = self._size + j
        if (c.ofs_begin() <= i) and (j < c.ofs_end()):
            b = c.ofs_begin()
            return c.buffer()[i - b:j - b]
        else:
            l = j - i                 # total length
            ofs = i
            # It's fastest to keep tokens and join later, especially in py3, which was 7 times slower
            # in the previous iteration of this code
            md = list()
            while l:
                c.use_region(ofs, l)
                assert c.is_valid()
                d = c.buffer()[:l]
                ofs += len(d)
                l -= len(d)
                # Make sure we don't keep references, as c.use_region() might attempt to free resources, but
                # can't unless we use pure bytes
                if hasattr(d, 'tobytes'):
                    d = d.tobytes()
                md.append(d)
            # END while there are bytes to read
            return b''.join(md)
        # END fast or slow path
    #{ Interface

    def begin_access(self, cursor=None, offset=0, size=sys.maxsize, flags=0):
        """Call this before the first use of this instance. The method was already
        called by the constructor in case sufficient information was provided.

        For more information no the parameters, see the __init__ method
        :param path: if cursor is None the existing one will be used.
        :return: True if the buffer can be used"""
        if cursor:
            self._c = cursor
        # END update our cursor

        # reuse existing cursors if possible
        if self._c is not None and self._c.is_associated():
            res = self._c.use_region(offset, size, flags).is_valid()
            if res:
                # if given size is too large or default, we computer a proper size
                # If its smaller, we assume the combination between offset and size
                # as chosen by the user is correct and use it !
                # If not, the user is in trouble.
                if size > self._c.file_size():
                    size = self._c.file_size() - offset
                # END handle size
                self._size = size
            # END set size
            return res
        # END use our cursor
        return False

    def end_access(self):
        """Call this method once you are done using the instance. It is automatically
        called on destruction, and should be called just in time to allow system
        resources to be freed.

        Once you called end_access, you must call begin access before reusing this instance!"""
        self._size = 0
        if self._c is not None:
            self._c.unuse_region()
        # END unuse region

    def cursor(self):
        """:return: the currently set cursor which provides access to the data"""
        return self._c

    #}END interface



"""
OpenTelemetry tracing integration for Salt.

This module exposes a small, opinionated wrapper over the OpenTelemetry SDK
so that the rest of the codebase can call ``start_span``, ``inject`` and
``extract`` unconditionally regardless of whether tracing is enabled.

When ``opts['tracing']['enabled']`` is false (the default), every public
function short-circuits and ``start_span`` returns a :class:`_NoopSpan`.  No
spans are created, no exporter is initialised and no background threads are
started.

The carrier format on the wire is W3C TraceContext: a ``traceparent`` (and
optional ``tracestate``) string injected into the appropriate dict / header
/ env var by the caller.

The provider is rebuilt per-PID.  ``BatchSpanProcessor`` runs a background
thread that is not preserved across ``fork``, so every public entry point
calls :func:`_ensure_tracer` which detects a PID change and rebuilds the
provider, processor and exporter in the child.

Configuration lives in ``opts['tracing']``::

    tracing:
      enabled: false
      exporter: otlp-http           # otlp-http | otlp-grpc | console
      endpoint: ""                  # OTel SDK default (4318/v1/traces for http, 4317 for grpc)
      service_name: ""              # auto-derived when empty
      sampler: parent_based         # parent_based | always_on | always_off | trace_id_ratio
      sampler_arg: 1.0
      resource_attributes: {}
      insecure: true
      headers: {}

The default ``otlp-http`` exporter is pure-Python and ships in salt's base
requirements.  The ``otlp-grpc`` exporter is opt-in: install
``opentelemetry-exporter-otlp-proto-grpc`` separately to use it.
"""

import atexit
import contextlib
import logging
import os
import threading

log = logging.getLogger(__name__)

_INSTRUMENTATION_NAME = "salt"

# OpenTelemetry is optional.  It is not shipped in the salt-ssh thin
# tarball, may be absent from older installed onedirs that the upgrade /
# downgrade tests still exercise, and may be intentionally uninstalled by
# operators who want a minimal footprint.  When opentelemetry is missing,
# every public function in this module short-circuits to a no-op, exactly
# as if ``opts['tracing']['enabled']`` were false.
try:
    from opentelemetry import context as otel_context
    from opentelemetry import trace
    from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
        OTLPSpanExporter as _OTLPSpanExporterHTTP,
    )
    from opentelemetry.sdk.resources import Resource
    from opentelemetry.sdk.trace import TracerProvider
    from opentelemetry.sdk.trace.export import BatchSpanProcessor, ConsoleSpanExporter
    from opentelemetry.sdk.trace.sampling import (
        ALWAYS_OFF,
        ALWAYS_ON,
        ParentBased,
        TraceIdRatioBased,
    )
    from opentelemetry.trace.propagation.tracecontext import (
        TraceContextTextMapPropagator,
    )

    _OTEL_AVAILABLE = True
    SpanKind = trace.SpanKind
except ImportError:  # pragma: no cover - exercised when opentelemetry is absent
    _OTEL_AVAILABLE = False
    otel_context = None  # type: ignore[assignment]
    trace = None  # type: ignore[assignment]
    _OTLPSpanExporterHTTP = None  # type: ignore[assignment]
    Resource = None  # type: ignore[assignment]
    TracerProvider = None  # type: ignore[assignment]
    BatchSpanProcessor = None  # type: ignore[assignment]
    ConsoleSpanExporter = None  # type: ignore[assignment]
    ALWAYS_OFF = ALWAYS_ON = ParentBased = TraceIdRatioBased = None  # type: ignore[assignment]
    TraceContextTextMapPropagator = None  # type: ignore[assignment]

    class _SpanKindStub:
        """Duck-typed ``trace.SpanKind`` used when opentelemetry is missing."""

        INTERNAL = "INTERNAL"
        SERVER = "SERVER"
        CLIENT = "CLIENT"
        PRODUCER = "PRODUCER"
        CONSUMER = "CONSUMER"

    SpanKind = _SpanKindStub()  # type: ignore[assignment]


_lock = threading.Lock()
_last_pid = None
_provider = None
_tracer = None
_cached_opts = None
_propagator = TraceContextTextMapPropagator() if _OTEL_AVAILABLE else None
_atexit_registered = False


class _InvalidSpanContext:
    """Minimal stand-in for ``trace.INVALID_SPAN_CONTEXT`` when otel is absent."""

    trace_id = 0
    span_id = 0
    is_valid = False
    is_remote = False
    trace_flags = 0
    trace_state = None


_INVALID_SPAN_CONTEXT_FALLBACK = _InvalidSpanContext()


class _NoopSpan:
    """Stand-in returned when tracing is disabled or opentelemetry is absent."""

    is_recording = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        return False

    def set_attribute(self, key, value):  # noqa: ARG002
        return None

    def set_attributes(self, attributes):  # noqa: ARG002
        return None

    def add_event(self, name, attributes=None, timestamp=None):  # noqa: ARG002
        return None

    def record_exception(self, exception, attributes=None):  # noqa: ARG002
        return None

    def set_status(self, status, description=None):  # noqa: ARG002
        return None

    def update_name(self, name):  # noqa: ARG002
        return None

    def end(self, end_time=None):  # noqa: ARG002
        return None

    def get_span_context(self):
        if _OTEL_AVAILABLE:
            return trace.INVALID_SPAN_CONTEXT
        return _INVALID_SPAN_CONTEXT_FALLBACK


_NOOP_SPAN = _NoopSpan()


def is_enabled():
    """Return True if tracing is configured and enabled."""
    if not _OTEL_AVAILABLE:
        return False
    return bool(_cached_opts and _cached_opts.get("enabled"))


def configure(opts):
    """
    Initialise tracing for this process.

    ``opts`` is the full Salt opts dict; the ``tracing`` block is read out of
    it.  Safe to call multiple times; the provider is rebuilt only when the
    PID changes or the cached configuration is empty.

    When tracing is disabled — or when opentelemetry is not installed —
    this is a cheap no-op that just caches the opts so that subsequent
    calls in fork children can pick up the same setting.
    """
    global _cached_opts, _atexit_registered
    tracing_opts = (opts or {}).get("tracing") or {}
    _cached_opts = dict(tracing_opts)
    _cached_opts.setdefault("service_name", _default_service_name(opts))
    if not _OTEL_AVAILABLE:
        if _cached_opts.get("enabled"):
            log.warning(
                "tracing.enabled is true but opentelemetry is not installed; "
                "tracing remains disabled in this process."
            )
        return
    if not _atexit_registered:
        atexit.register(shutdown)
        _atexit_registered = True
    if not _cached_opts.get("enabled"):
        log.debug(
            "tracing.configure called but tracing.enabled is false (pid=%d, service=%s)",
            os.getpid(),
            _cached_opts.get("service_name"),
        )
        return
    log.info(
        "Enabling OpenTelemetry tracing (pid=%d, service=%s, exporter=%s, endpoint=%s)",
        os.getpid(),
        _cached_opts.get("service_name"),
        _cached_opts.get("exporter"),
        _cached_opts.get("endpoint") or "<default>",
    )
    _ensure_tracer()


def shutdown():
    """Flush and tear down the active provider."""
    global _provider, _tracer, _last_pid
    with _lock:
        provider = _provider
        _provider = None
        _tracer = None
        _last_pid = None
    if provider is not None:
        try:
            provider.shutdown()
        except Exception:  # pylint: disable=broad-except
            log.debug("tracing provider shutdown raised", exc_info=True)


def start_span(name, *, kind=None, attributes=None, links=None, context=None):
    """
    Open a span as a context manager.

    Returns a no-op context manager when tracing is disabled.  ``context``
    is an opaque value previously returned from :func:`extract` and is used
    to link the new span as a child of a remote parent.
    """
    if not is_enabled():
        return _NOOP_SPAN
    _ensure_tracer()
    if _tracer is None:
        return _NOOP_SPAN
    if context is not None:
        return _start_with_context(name, context, kind, attributes, links)
    return _tracer.start_as_current_span(
        name,
        kind=kind or trace.SpanKind.INTERNAL,
        attributes=attributes,
        links=links,
    )


@contextlib.contextmanager
def _start_with_context(name, ctx, kind, attributes, links):
    token = otel_context.attach(ctx)
    try:
        with _tracer.start_as_current_span(
            name,
            kind=kind or trace.SpanKind.INTERNAL,
            attributes=attributes,
            links=links,
        ) as span:
            yield span
    finally:
        otel_context.detach(token)


def current_span():
    """Return the currently active span, or a :class:`_NoopSpan`."""
    if not is_enabled():
        return _NOOP_SPAN
    return trace.get_current_span()


def set_attribute(key, value):
    """Set an attribute on the current span (no-op when disabled)."""
    if not is_enabled():
        return
    span = trace.get_current_span()
    if span is not None and span.is_recording():
        span.set_attribute(key, value)


def record_exception(exc):
    """Record an exception on the current span (no-op when disabled)."""
    if not is_enabled():
        return
    span = trace.get_current_span()
    if span is not None and span.is_recording():
        span.record_exception(exc)


def inject(carrier):
    """
    Inject the current trace context into ``carrier``.

    ``carrier`` is any ``MutableMapping`` (e.g. the inner ``load`` dict of a
    Salt request, an event ``data`` dict, an HTTP header mapping or an env
    var dict).  When there is no recording span — or when opentelemetry is
    not installed — this is a no-op so the on-the-wire payload is not
    bloated with empty headers.
    """
    if not is_enabled() or _propagator is None:
        return
    span = trace.get_current_span()
    if span is None or not span.is_recording():
        return
    _propagator.inject(carrier)


def extract(carrier):
    """
    Extract a context from ``carrier``.

    Returns an opaque context object suitable for passing to
    :func:`start_span` as ``context=...``, or ``None`` when no context was
    found, tracing is disabled, or opentelemetry is not installed.
    """
    if not is_enabled() or not carrier or _propagator is None:
        return None
    ctx = _propagator.extract(carrier)
    if ctx is otel_context.Context():
        return None
    return ctx


def _ensure_tracer():
    global _last_pid  # pylint: disable=global-statement
    pid = os.getpid()
    if _last_pid == pid and _provider is not None:
        return
    with _lock:
        if _last_pid == pid and _provider is not None:
            return
        if _cached_opts is None or not _cached_opts.get("enabled"):
            return
        _build_provider()
        _last_pid = pid


def _build_provider():
    global _provider, _tracer
    opts = _cached_opts or {}
    resource = _build_resource(opts)
    sampler = _build_sampler(opts)
    provider = TracerProvider(resource=resource, sampler=sampler)
    exporter = _build_exporter(opts)
    if exporter is not None:
        provider.add_span_processor(BatchSpanProcessor(exporter))
    _provider = provider
    _tracer = provider.get_tracer(_INSTRUMENTATION_NAME)


def _build_resource(opts):
    attrs = {"service.name": opts.get("service_name") or "salt"}
    extra = opts.get("resource_attributes") or {}
    if isinstance(extra, dict):
        attrs.update(extra)
    return Resource.create(attrs)


def _build_sampler(opts):
    name = (opts.get("sampler") or "parent_based").lower()
    arg = opts.get("sampler_arg", 1.0)
    if name == "always_on":
        return ALWAYS_ON
    if name == "always_off":
        return ALWAYS_OFF
    if name == "trace_id_ratio":
        return TraceIdRatioBased(float(arg))
    if name == "parent_based":
        try:
            ratio = float(arg)
        except (TypeError, ValueError):
            ratio = 1.0
        root = ALWAYS_ON if ratio >= 1.0 else TraceIdRatioBased(ratio)
        return ParentBased(root=root)
    log.warning(
        "Unknown tracing sampler %r; defaulting to parent_based+always_on", name
    )
    return ParentBased(root=ALWAYS_ON)


def _build_exporter(opts):
    name = (opts.get("exporter") or "otlp-http").lower()
    endpoint = opts.get("endpoint") or None
    headers = opts.get("headers") or None
    insecure = opts.get("insecure", True)
    try:
        if name == "console":
            return ConsoleSpanExporter()
        if name == "otlp-http":
            kwargs = {}
            if endpoint:
                kwargs["endpoint"] = endpoint
            if headers:
                kwargs["headers"] = headers
            return _OTLPSpanExporterHTTP(**kwargs)
        if name == "otlp-grpc":
            # The gRPC exporter pulls in grpcio which has no wheel for some
            # interpreter / platform combinations.  Import lazily so the
            # default HTTP path works even when grpc isn't installed.
            try:
                from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import (
                    OTLPSpanExporter as _OTLPSpanExporterGRPC,
                )
            except ImportError:
                log.error(
                    "opentelemetry-exporter-otlp-proto-grpc is not installed; "
                    "either install it or set tracing.exporter to 'otlp-http'."
                )
                return None
            kwargs = {"insecure": bool(insecure)}
            if endpoint:
                kwargs["endpoint"] = endpoint
            if headers:
                kwargs["headers"] = headers
            return _OTLPSpanExporterGRPC(**kwargs)
    except Exception:  # pylint: disable=broad-except
        log.exception("Failed to build tracing exporter %r", name)
        return None
    log.warning("Unknown tracing exporter %r; tracing will be a no-op", name)
    return None


def _default_service_name(opts):
    if not opts:
        return "salt"
    role = opts.get("__role")
    if role == "master":
        return "salt-master"
    if role == "minion":
        minion_id = opts.get("id") or ""
        return f"salt-minion-{minion_id}" if minion_id else "salt-minion"
    return "salt"



import io as StringIO
import re
import string
from typing import Dict, Iterable, List, Match, Optional, TextIO, Tuple

from .metrics_core import Metric
from .samples import Sample
from .validation import (
    _is_valid_legacy_metric_name, _validate_labelname, _validate_metric_name,
)


def text_string_to_metric_families(text: str) -> Iterable[Metric]:
    """Parse Prometheus text format from a unicode string.

    See text_fd_to_metric_families.
    """
    yield from text_fd_to_metric_families(StringIO.StringIO(text))


ESCAPE_SEQUENCES = {
    '\\\\': '\\',
    '\\n': '\n',
    '\\"': '"',
}


def replace_escape_sequence(match: Match[str]) -> str:
    return ESCAPE_SEQUENCES[match.group(0)]


HELP_ESCAPING_RE = re.compile(r'\\[\\n]')
ESCAPING_RE = re.compile(r'\\[\\n"]')


def _replace_help_escaping(s: str) -> str:
    return HELP_ESCAPING_RE.sub(replace_escape_sequence, s)


def _replace_escaping(s: str) -> str:
    return ESCAPING_RE.sub(replace_escape_sequence, s)


def _is_character_escaped(s: str, charpos: int) -> bool:
    num_bslashes = 0
    while (charpos > num_bslashes
           and s[charpos - 1 - num_bslashes] == '\\'):
        num_bslashes += 1
    return num_bslashes % 2 == 1


def parse_labels(labels_string: str, openmetrics: bool = False) -> Dict[str, str]:
    labels: Dict[str, str] = {}

    # Copy original labels
    sub_labels = labels_string.strip()
    if openmetrics and sub_labels and sub_labels[0] == ',':
        raise ValueError("leading comma: " + labels_string)
    try:
        # Process one label at a time
        while sub_labels:
            # The label name is before the equal, or if there's no equal, that's the
            # metric name.
            
            name_term, value_term, sub_labels = _next_term(sub_labels, openmetrics)
            if not value_term:
                if openmetrics:
                    raise ValueError("empty term in line: " + labels_string)
                continue
            
            label_name, quoted_name = _unquote_unescape(name_term)
                
            if not quoted_name and not _is_valid_legacy_metric_name(label_name):
                raise ValueError("unquoted UTF-8 metric name")
                
            # Check for missing quotes 
            if not value_term or value_term[0] != '"':
                raise ValueError

            # The first quote is guaranteed to be after the equal.
            # Make sure that the next unescaped quote is the last character.
            i = 1
            while i < len(value_term):
                i = value_term.index('"', i)
                if not _is_character_escaped(value_term[:i], i):
                    break
                i += 1
            # The label value is between the first and last quote
            quote_end = i + 1
            if quote_end != len(value_term):
                raise ValueError("unexpected text after quote: " + labels_string)

            label_value, _ = _unquote_unescape(value_term)
            if label_name == '__name__':
                _validate_metric_name(label_name)
            else:
                _validate_labelname(label_name)
            if label_name in labels:
                raise ValueError("invalid line, duplicate label name: " + labels_string)
            labels[label_name] = label_value
        return labels
    except ValueError:
        raise ValueError("Invalid labels: " + labels_string)
    

def _next_term(text: str, openmetrics: bool) -> Tuple[str, str, str]:
    """Extract the next comma-separated label term from the text. The results
    are stripped terms for the label name, label value, and then the remainder
    of the string including the final , or }.
    
    Raises ValueError if the term is empty and we're in openmetrics mode.
    """
    
    # There may be a leading comma, which is fine here.
    if text[0] == ',':
        text = text[1:]
        if not text:
            return "", "", ""
        if text[0] == ',':
            raise ValueError("multiple commas")

    splitpos = _next_unquoted_char(text, '=,}')
    if splitpos >= 0 and text[splitpos] == "=":
        labelname = text[:splitpos]
        text = text[splitpos + 1:]
        splitpos = _next_unquoted_char(text, ',}')
    else:
        labelname = "__name__"

    if splitpos == -1:
        splitpos = len(text)
    term = text[:splitpos]
    if not term and openmetrics:
        raise ValueError("empty term:", term)
    
    rest = text[splitpos:]
    return labelname, term.strip(), rest.strip()


def _next_unquoted_char(text: str, chs: Optional[str], startidx: int = 0) -> int:
    """Return position of next unquoted character in tuple, or -1 if not found.
    
    It is always assumed that the first character being checked is not already
    inside quotes.
    """
    in_quotes = False
    if chs is None:
        chs = string.whitespace

    for i, c in enumerate(text[startidx:]):
        if c == '"' and not _is_character_escaped(text, startidx + i):
            in_quotes = not in_quotes
        if not in_quotes:
            if c in chs:
                return startidx + i
    return -1


def _last_unquoted_char(text: str, chs: Optional[str]) -> int:
    """Return position of last unquoted character in list, or -1 if not found."""
    i = len(text) - 1
    in_quotes = False
    if chs is None:
        chs = string.whitespace
    while i > 0:
        if text[i] == '"' and not _is_character_escaped(text, i):
            in_quotes = not in_quotes
            
        if not in_quotes:
            if text[i] in chs:
                return i
        i -= 1
    return -1


def _split_quoted(text, separator, maxsplit=0):
    """Splits on split_ch similarly to strings.split, skipping separators if
    they are inside quotes.
    """

    tokens = ['']
    x = 0
    while x < len(text):
        split_pos = _next_unquoted_char(text, separator, x)
        if split_pos == -1:
            tokens[-1] = text[x:]
            x = len(text)
            continue
        # If the first character is the separator keep going. This happens when
        # there are double whitespace characters separating symbols.
        if split_pos == x:
            x += 1
            continue

        if maxsplit > 0 and len(tokens) > maxsplit:
            tokens[-1] = text[x:]
            break
        tokens[-1] = text[x:split_pos]
        x = split_pos + 1
        tokens.append('')
    return tokens


def _unquote_unescape(text):
    """Returns the string, and true if it was quoted."""
    if not text:
        return text, False
    quoted = False
    text = text.strip()
    if text[0] == '"':
        if len(text) == 1 or text[-1] != '"':
            raise ValueError("missing close quote")
        text = text[1:-1]
        quoted = True
    if "\\" in text:
        text = _replace_escaping(text)
    return text, quoted


# If we have multiple values only consider the first
def _parse_value_and_timestamp(s: str) -> Tuple[float, Optional[float]]:
    s = s.lstrip()
    separator = " "
    if separator not in s:
        separator = "\t"
    values = [value.strip() for value in s.split(separator) if value.strip()]
    if not values:
        return float(s), None
    value = _parse_value(values[0])
    timestamp = (_parse_value(values[-1]) / 1000) if len(values) > 1 else None
    return value, timestamp


def _parse_value(value):
    value = ''.join(value)
    if value != value.strip() or '_' in value:
        raise ValueError(f"Invalid value: {value!r}")
    try:
        return int(value)
    except ValueError:
        return float(value)
    

def _parse_sample(text):
    separator = " # "
    # Detect the labels in the text
    label_start = _next_unquoted_char(text, '{')
    if label_start == -1 or separator in text[:label_start]:
        # We don't have labels, but there could be an exemplar.
        name_end = _next_unquoted_char(text, ' \t')
        name = text[:name_end].strip()
        if not _is_valid_legacy_metric_name(name):
            raise ValueError("invalid metric name:" + text)
        # Parse the remaining text after the name
        remaining_text = text[name_end + 1:]
        value, timestamp = _parse_value_and_timestamp(remaining_text)
        return Sample(name, {}, value, timestamp)
    name = text[:label_start].strip()
    label_end = _next_unquoted_char(text[label_start:], '}') + label_start
    labels = parse_labels(text[label_start + 1:label_end], False)
    if not name:
        # Name might be in the labels
        if '__name__' not in labels:
            raise ValueError
        name = labels['__name__']
        del labels['__name__']
    elif '__name__' in labels:
        raise ValueError("metric name specified more than once")
    # Parsing labels succeeded, continue parsing the remaining text
    remaining_text = text[label_end + 1:]
    value, timestamp = _parse_value_and_timestamp(remaining_text)
    return Sample(name, labels, value, timestamp)


def text_fd_to_metric_families(fd: TextIO) -> Iterable[Metric]:
    """Parse Prometheus text format from a file descriptor.

    This is a laxer parser than the main Go parser,
    so successful parsing does not imply that the parsed
    text meets the specification.

    Yields Metric's.
    """
    name = ''
    documentation = ''
    typ = 'untyped'
    samples: List[Sample] = []
    allowed_names = []

    def build_metric(name: str, documentation: str, typ: str, samples: List[Sample]) -> Metric:
        # Munge counters into OpenMetrics representation
        # used internally.
        if typ == 'counter':
            if name.endswith('_total'):
                name = name[:-6]
            else:
                new_samples = []
                for s in samples:
                    new_samples.append(Sample(s[0] + '_total', *s[1:]))
                    samples = new_samples
        metric = Metric(name, documentation, typ)
        metric.samples = samples
        return metric

    for line in fd:
        line = line.strip()

        if line.startswith('#'):
            parts = _split_quoted(line, None, 3)
            if len(parts) < 2:
                continue
            candidate_name, quoted = '', False
            if len(parts) > 2:
                # Ignore comment tokens
                if parts[1] != 'TYPE' and parts[1] != 'HELP':
                    continue
                candidate_name, quoted = _unquote_unescape(parts[2])
                if not quoted and not _is_valid_legacy_metric_name(candidate_name):
                    raise ValueError
            if parts[1] == 'HELP':
                if candidate_name != name:
                    if name != '':
                        yield build_metric(name, documentation, typ, samples)
                    # New metric
                    name = candidate_name
                    typ = 'untyped'
                    samples = []
                    allowed_names = [candidate_name]
                if len(parts) == 4:
                    documentation = _replace_help_escaping(parts[3])
                else:
                    documentation = ''
            elif parts[1] == 'TYPE':
                if len(parts) < 4:
                    raise ValueError
                if candidate_name != name:
                    if name != '':
                        yield build_metric(name, documentation, typ, samples)
                    # New metric
                    name = candidate_name
                    documentation = ''
                    samples = []
                typ = parts[3]
                allowed_names = {
                    'counter': [''],
                    'gauge': [''],
                    'summary': ['_count', '_sum', ''],
                    'histogram': ['_count', '_sum', '_bucket'],
                }.get(typ, [''])
                allowed_names = [name + n for n in allowed_names]
        elif line == '':
            # Ignore blank lines
            pass
        else:
            sample = _parse_sample(line)
            if sample.name not in allowed_names:
                if name != '':
                    yield build_metric(name, documentation, typ, samples)
                # New metric, yield immediately as untyped singleton
                name = ''
                documentation = ''
                typ = 'untyped'
                samples = []
                allowed_names = []
                yield build_metric(sample[0], documentation, typ, [sample])
            else:
                samples.append(sample)

    if name != '':
        yield build_metric(name, documentation, typ, samples)



"""
Default values, to be imported elsewhere in Salt code

Do NOT, import any salt modules (salt.utils, salt.config, etc.) into this file,
as this may result in circular imports.
"""


class _Constant:
    """
    This class implements a way to create constants in python.

    NOTE:

      - This is not really a constant, ie, the `is` check will not work, you'll
        have to use `==`.
      - This class SHALL NOT be considered public API and might change or even
        go away at any given time.
    """

    __slots__ = ("name", "value")

    def __init__(self, name, value=None):
        self.name = name
        self.value = value

    def __hash__(self):
        return hash((self.name, self.value))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if self.name != other.name:
            return False
        return self.value == other.value

    def __get_state__(self):
        return {
            "name": self.name,
            "value": self.value,
        }

    def __set_state__(self, state):
        return self.__class__(state["name"], state["value"])

    def __repr__(self):
        if self.value:
            return f"<Constant.{self.name} value={self.value}>"
        return f"<Constant.{self.name}>"


# Default delimiter for multi-level traversal in targeting
DEFAULT_TARGET_DELIM = ":"


"""
Used in functions to define that a keyword default is not set.

It's used to differentiate from `None`, `True`, `False` which, in some
cases are proper defaults and are also proper values to pass.
"""
NOT_SET = _Constant("NOT_SET")



from __future__ import annotations

from virtualenv.seed.wheels.embed import get_embed_wheel

from .periodic_update import periodic_update
from .util import Version, Wheel, discover_wheels


def from_bundle(distribution, version, for_py_version, search_dirs, app_data, do_periodic_update, env):  # noqa: PLR0913
    """Load the bundled wheel to a cache directory."""
    of_version = Version.of_version(version)
    wheel = load_embed_wheel(app_data, distribution, for_py_version, of_version)

    if version != Version.embed:
        # 2. check if we have upgraded embed
        if app_data.can_update:
            per = do_periodic_update
            wheel = periodic_update(distribution, of_version, for_py_version, wheel, search_dirs, app_data, per, env)

        # 3. acquire from extra search dir
        found_wheel = from_dir(distribution, of_version, for_py_version, search_dirs)
        if found_wheel is not None and (wheel is None or found_wheel.version_tuple > wheel.version_tuple):
            wheel = found_wheel
    return wheel


def load_embed_wheel(app_data, distribution, for_py_version, version):
    wheel = get_embed_wheel(distribution, for_py_version)
    if wheel is not None:
        version_match = version == wheel.version
        if version is None or version_match:
            with app_data.ensure_extracted(wheel.path, lambda: app_data.house) as wheel_path:
                wheel = Wheel(wheel_path)
        else:  # if version does not match ignore
            wheel = None
    return wheel


def from_dir(distribution, version, for_py_version, directories):
    """Load a compatible wheel from a given folder."""
    for folder in directories:
        for wheel in discover_wheels(folder, distribution, version, for_py_version):
            return wheel
    return None


__all__ = [
    "from_bundle",
    "load_embed_wheel",
]



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2019, OpenTelemetry Authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class AnyValue(google.protobuf.message.Message):
    """Represents any type of attribute value. AnyValue may contain a
    primitive value such as a string or integer or it may contain an arbitrary nested
    object containing arrays, key-value lists and primitives.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STRING_VALUE_FIELD_NUMBER: builtins.int
    BOOL_VALUE_FIELD_NUMBER: builtins.int
    INT_VALUE_FIELD_NUMBER: builtins.int
    DOUBLE_VALUE_FIELD_NUMBER: builtins.int
    ARRAY_VALUE_FIELD_NUMBER: builtins.int
    KVLIST_VALUE_FIELD_NUMBER: builtins.int
    BYTES_VALUE_FIELD_NUMBER: builtins.int
    string_value: builtins.str
    bool_value: builtins.bool
    int_value: builtins.int
    double_value: builtins.float
    @property
    def array_value(self) -> global___ArrayValue: ...
    @property
    def kvlist_value(self) -> global___KeyValueList: ...
    bytes_value: builtins.bytes
    def __init__(
        self,
        *,
        string_value: builtins.str = ...,
        bool_value: builtins.bool = ...,
        int_value: builtins.int = ...,
        double_value: builtins.float = ...,
        array_value: global___ArrayValue | None = ...,
        kvlist_value: global___KeyValueList | None = ...,
        bytes_value: builtins.bytes = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["array_value", b"array_value", "bool_value", b"bool_value", "bytes_value", b"bytes_value", "double_value", b"double_value", "int_value", b"int_value", "kvlist_value", b"kvlist_value", "string_value", b"string_value", "value", b"value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["array_value", b"array_value", "bool_value", b"bool_value", "bytes_value", b"bytes_value", "double_value", b"double_value", "int_value", b"int_value", "kvlist_value", b"kvlist_value", "string_value", b"string_value", "value", b"value"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["value", b"value"]) -> typing_extensions.Literal["string_value", "bool_value", "int_value", "double_value", "array_value", "kvlist_value", "bytes_value"] | None: ...

global___AnyValue = AnyValue

@typing_extensions.final
class ArrayValue(google.protobuf.message.Message):
    """ArrayValue is a list of AnyValue messages. We need ArrayValue as a message
    since oneof in AnyValue does not allow repeated fields.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALUES_FIELD_NUMBER: builtins.int
    @property
    def values(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___AnyValue]:
        """Array of values. The array may be empty (contain 0 elements)."""
    def __init__(
        self,
        *,
        values: collections.abc.Iterable[global___AnyValue] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["values", b"values"]) -> None: ...

global___ArrayValue = ArrayValue

@typing_extensions.final
class KeyValueList(google.protobuf.message.Message):
    """KeyValueList is a list of KeyValue messages. We need KeyValueList as a message
    since `oneof` in AnyValue does not allow repeated fields. Everywhere else where we need
    a list of KeyValue messages (e.g. in Span) we use `repeated KeyValue` directly to
    avoid unnecessary extra wrapping (which slows down the protocol). The 2 approaches
    are semantically equivalent.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALUES_FIELD_NUMBER: builtins.int
    @property
    def values(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___KeyValue]:
        """A collection of key/value pairs of key-value pairs. The list may be empty (may
        contain 0 elements).

        The keys MUST be unique (it is not allowed to have more than one
        value with the same key).
        The behavior of software that receives duplicated keys can be unpredictable.
        """
    def __init__(
        self,
        *,
        values: collections.abc.Iterable[global___KeyValue] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["values", b"values"]) -> None: ...

global___KeyValueList = KeyValueList

@typing_extensions.final
class KeyValue(google.protobuf.message.Message):
    """Represents a key-value pair that is used to store Span attributes, Link
    attributes, etc.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KEY_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    key: builtins.str
    """The key name of the pair."""
    @property
    def value(self) -> global___AnyValue:
        """The value of the pair."""
    def __init__(
        self,
        *,
        key: builtins.str = ...,
        value: global___AnyValue | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["value", b"value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

global___KeyValue = KeyValue

@typing_extensions.final
class InstrumentationScope(google.protobuf.message.Message):
    """InstrumentationScope is a message representing the instrumentation scope information
    such as the fully qualified name and version.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    VERSION_FIELD_NUMBER: builtins.int
    ATTRIBUTES_FIELD_NUMBER: builtins.int
    DROPPED_ATTRIBUTES_COUNT_FIELD_NUMBER: builtins.int
    name: builtins.str
    """A name denoting the Instrumentation scope.
    An empty instrumentation scope name means the name is unknown.
    """
    version: builtins.str
    """Defines the version of the instrumentation scope.
    An empty instrumentation scope version means the version is unknown.
    """
    @property
    def attributes(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___KeyValue]:
        """Additional attributes that describe the scope. [Optional].
        Attribute keys MUST be unique (it is not allowed to have more than one
        attribute with the same key).
        The behavior of software that receives duplicated keys can be unpredictable.
        """
    dropped_attributes_count: builtins.int
    """The number of attributes that were discarded. Attributes
    can be discarded because their keys are too long or because there are too many
    attributes. If this value is 0, then no attributes were dropped.
    """
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        version: builtins.str = ...,
        attributes: collections.abc.Iterable[global___KeyValue] | None = ...,
        dropped_attributes_count: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["attributes", b"attributes", "dropped_attributes_count", b"dropped_attributes_count", "name", b"name", "version", b"version"]) -> None: ...

global___InstrumentationScope = InstrumentationScope

@typing_extensions.final
class EntityRef(google.protobuf.message.Message):
    """A reference to an Entity.
    Entity represents an object of interest associated with produced telemetry: e.g spans, metrics, profiles, or logs.

    Status: [Development]
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SCHEMA_URL_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    ID_KEYS_FIELD_NUMBER: builtins.int
    DESCRIPTION_KEYS_FIELD_NUMBER: builtins.int
    schema_url: builtins.str
    """The Schema URL, if known. This is the identifier of the Schema that the entity data
    is recorded in. To learn more about Schema URL see
    https://opentelemetry.io/docs/specs/otel/schemas/#schema-url

    This schema_url applies to the data in this message and to the Resource attributes
    referenced by id_keys and description_keys.
    TODO: discuss if we are happy with this somewhat complicated definition of what
    the schema_url applies to.

    This field obsoletes the schema_url field in ResourceMetrics/ResourceSpans/ResourceLogs.
    """
    type: builtins.str
    """Defines the type of the entity. MUST not change during the lifetime of the entity.
    For example: "service" or "host". This field is required and MUST not be empty
    for valid entities.
    """
    @property
    def id_keys(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """Attribute Keys that identify the entity.
        MUST not change during the lifetime of the entity. The Id must contain at least one attribute.
        These keys MUST exist in the containing {message}.attributes.
        """
    @property
    def description_keys(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """Descriptive (non-identifying) attribute keys of the entity.
        MAY change over the lifetime of the entity. MAY be empty.
        These attribute keys are not part of entity's identity.
        These keys MUST exist in the containing {message}.attributes.
        """
    def __init__(
        self,
        *,
        schema_url: builtins.str = ...,
        type: builtins.str = ...,
        id_keys: collections.abc.Iterable[builtins.str] | None = ...,
        description_keys: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["description_keys", b"description_keys", "id_keys", b"id_keys", "schema_url", b"schema_url", "type", b"type"]) -> None: ...

global___EntityRef = EntityRef



"""
Used to manage the outputter system. This package is the modular system used
for managing outputters.
"""

import errno
import io
import logging
import os
import re
import sys
import traceback

import salt.loader
import salt.utils.files
import salt.utils.platform
import salt.utils.secret
import salt.utils.stringutils

# Are you really sure !!!
# dealing with unicode is not as simple as setting defaultencoding
# which can break other python modules imported by salt in bad ways...
# reloading sys is not either a good idea...
# reload(sys)
# sys.setdefaultencoding('utf-8')

log = logging.getLogger(__name__)


def try_printout(data, out, opts, **kwargs):
    """
    Safely get the string to print out, try the configured outputter, then
    fall back to nested and then to raw
    """
    data = salt.utils.secret.mask_output(data)
    try:
        printout = get_printout(out, opts)(data, **kwargs)
        if printout is not None:
            return printout.rstrip()
    except (KeyError, AttributeError, TypeError):
        log.debug(traceback.format_exc())
        try:
            printout = get_printout("nested", opts)(data, **kwargs)
            if printout is not None:
                return printout.rstrip()
        except (KeyError, AttributeError, TypeError):
            log.error("Nested output failed: ", exc_info=True)
            printout = get_printout("raw", opts)(data, **kwargs)
            if printout is not None:
                return printout.rstrip()


def get_progress(opts, out, progress):
    """
    Get the progress bar from the given outputter
    """
    return salt.loader.raw_mod(opts, out, "rawmodule", mod="output")[
        f"{out}.progress_iter"
    ](progress)


def update_progress(opts, progress, progress_iter, out):
    """
    Update the progress iterator for the given outputter
    """
    # Look up the outputter
    try:
        progress_outputter = salt.loader.outputters(opts)[out]
    except KeyError:  # Outputter is not loaded
        log.warning("Progress outputter not available.")
        return False
    progress_outputter(progress, progress_iter)


def progress_end(progress_iter):
    try:
        progress_iter.stop()
    except Exception:  # pylint: disable=broad-except
        pass
    return None


def display_output(data, out=None, opts=None, **kwargs):
    """
    Print the passed data using the desired output
    """
    if opts is None:
        opts = {}
    display_data = try_printout(data, out, opts, **kwargs)

    output_filename = opts.get("output_file", None)
    log.trace("data = %s", data)
    try:
        # output filename can be either '' or None
        if output_filename:
            if not hasattr(output_filename, "write"):
                # pylint: disable=resource-leakage
                ofh = salt.utils.files.fopen(output_filename, "a")
                # pylint: enable=resource-leakage
                fh_opened = True
            else:
                # Filehandle/file-like object
                ofh = output_filename
                fh_opened = False

            try:
                fdata = display_data
                if isinstance(fdata, str):
                    try:
                        fdata = fdata.encode("utf-8")
                    except (UnicodeDecodeError, UnicodeEncodeError):
                        # try to let the stream write
                        # even if we didn't encode it
                        pass
                if fdata:
                    ofh.write(salt.utils.stringutils.to_str(fdata))
                    ofh.write("\n")
            finally:
                if fh_opened:
                    ofh.close()
            return
        if display_data:
            salt.utils.stringutils.print_cli(display_data)
    except OSError as exc:
        # Only raise if it's NOT a broken pipe
        if exc.errno != errno.EPIPE:
            raise


def get_printout(out, opts=None, **kwargs):
    """
    Return a printer function
    """
    if opts is None:
        opts = {}

    if "output" in opts and opts["output"] != "highstate":
        # new --out option, but don't choke when using --out=highstate at CLI
        # See Issue #29796 for more information.
        out = opts["output"]

    # Handle setting the output when --static is passed.
    if not out and opts.get("static"):
        if opts.get("output"):
            out = opts["output"]
        elif opts.get("fun", "").split(".")[0] == "state":
            # --static doesn't have an output set at this point, but if we're
            # running a state function and "out" hasn't already been set, we
            # should set the out variable to "highstate". Otherwise state runs
            # are set to "nested" below. See Issue #44556 for more information.
            out = "highstate"

    if out == "text":
        out = "txt"
    elif out is None or out == "":
        out = "nested"
    if opts.get("progress", False):
        out = "progress"

    opts.update(kwargs)
    if "color" not in opts:

        def is_pipe():
            """
            Check if sys.stdout is a pipe or not
            """
            try:
                fileno = sys.stdout.fileno()
            except (AttributeError, io.UnsupportedOperation):
                fileno = -1  # sys.stdout is StringIO or fake
            return not os.isatty(fileno)

        if opts.get("force_color", False):
            opts["color"] = True
        elif (
            opts.get("no_color", False) or is_pipe() or salt.utils.platform.is_windows()
        ):
            opts["color"] = False
        else:
            opts["color"] = True
    else:
        if opts.get("force_color", False):
            opts["color"] = True
        elif opts.get("no_color", False) or salt.utils.platform.is_windows():
            opts["color"] = False
        else:
            pass

    outputters = salt.loader.outputters(opts)
    if out not in outputters:
        # Since the grains outputter was removed we don't need to fire this
        # error when old minions are asking for it
        if out != "grains":
            log.error(
                "Invalid outputter %s specified, fall back to nested",
                out,
            )
        return outputters["nested"]
    return outputters[out]


def out_format(data, out, opts=None, **kwargs):
    """
    Return the formatted outputter string for the passed data
    """
    return try_printout(data, out, opts, **kwargs)


def string_format(data, out, opts=None, **kwargs):
    """
    Return the formatted outputter string, removing the ANSI escape sequences.
    """
    raw_output = try_printout(data, out, opts, **kwargs)
    ansi_escape = re.compile(r"\x1b[^m]*m")
    return ansi_escape.sub("", raw_output)


def html_format(data, out, opts=None, **kwargs):
    """
    Return the formatted string as HTML.
    """
    ansi_escaped_string = string_format(data, out, opts, **kwargs)
    return ansi_escaped_string.replace(" ", "&nbsp;").replace("\n", "<br />")


def strip_esc_sequence(txt):
    """
    Replace ESC (ASCII 27/Oct 33) to prevent unsafe strings
    from writing their own terminal manipulation commands
    """
    if isinstance(txt, str):
        return txt.replace("\033", "?")
    else:
        return txt



"""
Output adapters for batch execution.

Each adapter is called after ``progress_batch()`` returns to report
what happened.  The adapter chosen depends on the execution mode:

- ``CLIOutput`` — prints to stdout (Phase 2: today's sync batch behavior).
- ``EventOutput`` — fires events on the master event bus (async mode,
  BatchManager).
- ``SilentOutput`` — no output; job cache only (programmatic use).

This module also owns the batch event tag vocabulary.  Nothing else
in the codebase should format a ``salt/batch/*`` tag by hand.
"""

import logging
import time

import salt.output
import salt.utils.stringutils

log = logging.getLogger(__name__)


# ---------------------------------------------------------------------------
# Event tag vocabulary.  Nothing else in the codebase should format a
# ``salt/batch/*`` tag by hand; go through the helpers below.
# ---------------------------------------------------------------------------

_BATCH_NEW = "salt/batch/{jid}/new"
_BATCH_PROGRESS = "salt/batch/{jid}/progress"
_BATCH_COMPLETE = "salt/batch/{jid}/complete"
_BATCH_HALTED = "salt/batch/{jid}/halted"
_BATCH_RECOVER = "salt/batch/{jid}/recover"


def tag_new(jid):
    """Tag fired when a new async batch is registered."""
    return _BATCH_NEW.format(jid=jid)


def tag_progress(jid):
    """Tag fired on each state change and on the idle heartbeat."""
    return _BATCH_PROGRESS.format(jid=jid)


def tag_complete(jid):
    """Tag fired when all minions are accounted for and no halt occurred."""
    return _BATCH_COMPLETE.format(jid=jid)


def tag_halted(jid):
    """Tag fired on abnormal termination (failhard / stop / corrupt / stale)."""
    return _BATCH_HALTED.format(jid=jid)


def tag_recover(jid):
    """Tag fired by Maintenance when a batch appears stale."""
    return _BATCH_RECOVER.format(jid=jid)


# ---------------------------------------------------------------------------
# Payload builders — keep event bodies consistent between drivers.
# ---------------------------------------------------------------------------


def _base_payload(state):
    """Fields present in every batch event."""
    return {
        "jid": state.get("jid"),
        "fun": state.get("fun"),
        "tgt": state.get("tgt"),
        "tgt_type": state.get("tgt_type"),
        "user": state.get("user"),
    }


def new_payload(state):
    """Payload for ``salt/batch/<jid>/new``."""
    payload = _base_payload(state)
    payload.update(
        {
            "total_minions": len(state.get("all_minions", [])),
            "batch_size": state.get("batch_size"),
            "driver": state.get("driver"),
            "created": state.get("created"),
        }
    )
    return payload


def progress_payload(state, iteration=None):
    """Payload for ``salt/batch/<jid>/progress``."""
    payload = _base_payload(state)
    payload.update(
        {
            "total": len(state.get("all_minions", [])),
            "completed": len(state.get("done", {})),
            "active": len(state.get("active", {})),
            "pending": len(state.get("pending", [])),
            "failed": len(state.get("failed", {})),
            "iter": iteration,
            "last_progress": state.get("last_progress"),
        }
    )
    return payload


def complete_payload(state, now=None):
    """Payload for ``salt/batch/<jid>/complete``."""
    if now is None:
        now = time.time()
    payload = _base_payload(state)
    payload.update(
        {
            "completed": len(state.get("done", {})),
            "failed": len(state.get("failed", {})),
            "total_minions": len(state.get("all_minions", [])),
            "duration": now - state.get("created", now),
        }
    )
    return payload


def halted_payload(state, now=None):
    """Payload for ``salt/batch/<jid>/halted``."""
    if now is None:
        now = time.time()
    payload = _base_payload(state)
    payload.update(
        {
            "reason": state.get("halted_reason"),
            "completed": len(state.get("done", {})),
            "failed": len(state.get("failed", {})),
            "abandoned_active": sorted(state.get("active", {}).keys()),
            "abandoned_pending": list(state.get("pending", [])),
            "duration": now - state.get("created", now),
        }
    )
    return payload


def recover_payload(state, age_seconds, now=None):
    """Payload for ``salt/batch/<jid>/recover``."""
    payload = _base_payload(state)
    payload.update(
        {
            "reason": "stale",
            "age_seconds": age_seconds,
            "driver": state.get("driver"),
        }
    )
    return payload


# ---------------------------------------------------------------------------
# Adapters
# ---------------------------------------------------------------------------


class CLIOutput:
    """
    Print batch progress to stdout.

    Replaces the stdout formatting previously inlined in
    ``salt/cli/batch.py:Batch.run()``.  Every hook honors ``quiet``
    — if it's truthy nothing is printed.  The observable output of a
    batch run with the refactored ``Batch.run()`` is byte-identical
    to the pre-Phase-2 implementation.
    """

    def __init__(self, opts, quiet=False):
        self.opts = opts
        self.quiet = quiet

    def on_minion_down(self, minion_id, **kwargs):
        """Announce a minion that failed the initial ping."""
        if self.quiet:
            return
        salt.utils.stringutils.print_cli(
            f"Minion {minion_id} did not respond. No job will be sent."
        )

    def on_batch_start(self, minion_ids, **kwargs):
        """Announce the minion list for a new sub-batch."""
        if self.quiet:
            return
        salt.utils.stringutils.print_cli(f"\nExecuting run on {sorted(minion_ids)}\n")

    def on_minion_return(self, minion_id, data, **kwargs):
        """
        Render a minion return via ``salt.output.display_output``.

        ``data`` is the minion's return dict as produced by
        ``LocalClient.cmd_iter_no_block`` — ``{"ret": ..., "retcode": ..., ...}``.
        The display shape matches the legacy sync formatter: the
        ``ret`` key is re-mapped under the minion id and ``out`` (if
        present) selects the outputter.
        """
        if self.quiet:
            return
        reshaped = dict(data)
        if "ret" in reshaped:
            reshaped[minion_id] = reshaped.pop("ret")
        out = reshaped.pop("out", None)
        salt.output.display_output(reshaped, out, self.opts)

    def on_minion_failed(self, minion_id, **kwargs):
        """Announce a minion whose return carried ``failed: True``."""
        if self.quiet:
            return
        salt.utils.stringutils.print_cli(
            f"Minion '{minion_id}' failed to respond to job sent"
        )

    def on_minion_timeout(self, minion_id, **kwargs):
        """No-op in CLI mode: timeouts surface through empty returns."""

    def on_batch_done(self, state, **kwargs):
        """No-op in CLI mode: the caller prints summaries at top level."""


class EventOutput:
    """
    Fire events on the master event bus for batch progress.

    Used by the BatchManager process in async mode.  Individual minion
    returns already fire ``salt/job/<jid>/ret/<minion>`` via the normal
    return path; this adapter fires higher-level batch lifecycle events
    only.

    Callers are expected to hold a reference to a
    ``salt.utils.event.MasterEvent`` (or equivalent) and hand it in at
    construction time.
    """

    def __init__(self, opts, event):
        """
        :param dict opts: Salt opts dictionary.
        :param event: Master event bus handle with a
            ``fire_event(data, tag)`` method.
        """
        self.opts = opts
        self.event = event

    def on_batch_new(self, state):
        """Fire ``salt/batch/<jid>/new``."""
        self.event.fire_event(new_payload(state), tag_new(state["jid"]))

    def on_batch_progress(self, state, iteration=None):
        """Fire ``salt/batch/<jid>/progress``."""
        self.event.fire_event(
            progress_payload(state, iteration=iteration),
            tag_progress(state["jid"]),
        )

    def on_batch_complete(self, state, now=None):
        """Fire ``salt/batch/<jid>/complete``."""
        self.event.fire_event(
            complete_payload(state, now=now), tag_complete(state["jid"])
        )

    def on_batch_halted(self, state, now=None):
        """Fire ``salt/batch/<jid>/halted``."""
        self.event.fire_event(halted_payload(state, now=now), tag_halted(state["jid"]))

    def on_batch_recover(self, state, age_seconds, now=None):
        """Fire ``salt/batch/<jid>/recover`` from the Maintenance safety net."""
        self.event.fire_event(
            recover_payload(state, age_seconds, now=now),
            tag_recover(state["jid"]),
        )

    # Per-minion hooks — kept as no-ops so the same adapter shape works
    # for drivers that want to surface per-minion detail in the future.
    # Per-minion returns are already on the bus via salt/job/*/ret/*.

    def on_batch_start(self, minion_ids, **kwargs):
        """No-op in Phase 1 — sub-batch publishes already fire ``salt/job/<jid>/new``."""

    def on_minion_return(self, minion_id, data, **kwargs):
        """No-op; minion returns are already on the bus via salt/job/*."""

    def on_minion_timeout(self, minion_id, **kwargs):
        """No-op; timeout visibility lives in the progress payload."""

    def on_batch_done(self, state, now=None, **kwargs):
        """
        Dispatch to ``on_batch_complete`` or ``on_batch_halted`` based
        on ``state["halted"]``.  Convenience wrapper so drivers don't
        have to branch.
        """
        if state.get("halted"):
            self.on_batch_halted(state, now=now)
        else:
            self.on_batch_complete(state, now=now)


class SilentOutput:
    """
    No output.  Job cache only.

    Used for programmatic batch execution where the caller doesn't
    need real-time reporting.  Every hook is a no-op.
    """

    def on_batch_new(self, state, **kwargs):
        """No-op."""

    def on_batch_progress(self, state, **kwargs):
        """No-op."""

    def on_batch_complete(self, state, **kwargs):
        """No-op."""

    def on_batch_halted(self, state, **kwargs):
        """No-op."""

    def on_batch_recover(self, state, **kwargs):
        """No-op."""

    def on_batch_start(self, minion_ids, **kwargs):
        """No-op."""

    def on_minion_return(self, minion_id, data, **kwargs):
        """No-op."""

    def on_minion_timeout(self, minion_id, **kwargs):
        """No-op."""

    def on_batch_done(self, state, **kwargs):
        """No-op."""



import math
from typing import Union

INF = float("inf")
MINUS_INF = float("-inf")
NaN = float("NaN")


def floatToGoString(d):
    d = float(d)
    if d == INF:
        return '+Inf'
    elif d == MINUS_INF:
        return '-Inf'
    elif math.isnan(d):
        return 'NaN'
    else:
        s = repr(d)
        dot = s.find('.')
        # Go switches to exponents sooner than Python.
        # We only need to care about positive values for le/quantile.
        if d > 0 and dot > 6:
            mantissa = f'{s[0]}.{s[1:dot]}{s[dot + 1:]}'.rstrip('0.')
            return f'{mantissa}e+0{dot - 1}'
        return s


def parse_version(version_str: str) -> tuple[Union[int, str], ...]:
    version: list[Union[int, str]] = []
    for part in version_str.split('.'):
        try:
            version.append(int(part))
        except ValueError:
            version.append(part)

    return tuple(version)



"""
Module for managing timezone on POSIX-like systems.
"""

import errno
import filecmp
import logging
import os
import re
import string

import salt.utils.files
import salt.utils.hashutils
import salt.utils.itertools
import salt.utils.path
import salt.utils.platform
import salt.utils.stringutils
from salt.config import DEFAULT_HASH_TYPE
from salt.exceptions import CommandExecutionError, SaltInvocationError

log = logging.getLogger(__name__)

__virtualname__ = "timezone"


def __virtual__():
    """
    Only work on POSIX-like systems
    """
    if salt.utils.platform.is_windows():
        return (
            False,
            "The timezone execution module failed to load: "
            "win_timezone.py should replace this module on Windows."
            "There was a problem loading win_timezone.py.",
        )

    if salt.utils.platform.is_darwin():
        return (
            False,
            "The timezone execution module failed to load: "
            "mac_timezone.py should replace this module on macOS."
            "There was a problem loading mac_timezone.py.",
        )

    return __virtualname__


def _timedatectl():
    """
    get the output of timedatectl
    """
    ret = __salt__["cmd.run_all"]("timedatectl", python_shell=False)

    if ret["retcode"] != 0:
        msg = "timedatectl failed: {}".format(ret["stderr"])
        raise CommandExecutionError(msg)

    return ret


def _get_zone_solaris():
    tzfile = "/etc/TIMEZONE"
    with salt.utils.files.fopen(tzfile, "r") as fp_:
        for line in fp_:
            line = salt.utils.stringutils.to_unicode(line)
            if "TZ=" in line:
                zonepart = line.rstrip("\n").split("=")[-1]
                return zonepart.strip("'\"") or "UTC"
    raise CommandExecutionError("Unable to get timezone from " + tzfile)


def _get_adjtime_timezone():
    """
    Return the timezone in /etc/adjtime of the system clock
    """
    adjtime_file = "/etc/adjtime"
    if os.path.exists(adjtime_file):
        cmd = ["tail", "-n", "1", adjtime_file]
        return __salt__["cmd.run"](cmd, python_shell=False)
    elif os.path.exists("/dev/rtc"):
        raise CommandExecutionError(
            "Unable to get hwclock timezone from " + adjtime_file
        )
    else:
        # There is no RTC.
        return None


def _get_zone_sysconfig():
    tzfile = "/etc/sysconfig/clock"
    with salt.utils.files.fopen(tzfile, "r") as fp_:
        for line in fp_:
            line = salt.utils.stringutils.to_unicode(line)
            if re.match(r"^\s*#", line):
                continue
            if "ZONE" in line and "=" in line:
                zonepart = line.rstrip("\n").split("=")[-1]
                return zonepart.strip("'\"") or "UTC"
    raise CommandExecutionError("Unable to get timezone from " + tzfile)


def _get_zone_etc_localtime():
    tzfile = _get_localtime_path()
    tzdir = "/usr/share/zoneinfo/"
    tzdir_len = len(tzdir)
    try:
        olson_name = os.path.normpath(os.path.join("/etc", os.readlink(tzfile)))
        if olson_name.startswith(tzdir):
            return olson_name[tzdir_len:]
    except OSError as exc:
        if exc.errno == errno.ENOENT:
            if "FreeBSD" in __grains__["os_family"]:
                return get_zonecode()
            raise CommandExecutionError(tzfile + " does not exist")
        elif exc.errno == errno.EINVAL:
            if "FreeBSD" in __grains__["os_family"]:
                return get_zonecode()
            log.warning(
                "%s is not a symbolic link. Attempting to match it to zoneinfo files.",
                tzfile,
            )
            # Regular file. Try to match the hash.
            hash_type = __opts__.get("hash_type", DEFAULT_HASH_TYPE)
            tzfile_hash = salt.utils.hashutils.get_hash(tzfile, hash_type)
            # Not a link, just a copy of the tzdata file
            for root, dirs, files in salt.utils.path.os_walk(tzdir):
                for filename in files:
                    full_path = os.path.join(root, filename)
                    olson_name = full_path[tzdir_len:]
                    if olson_name[0] in string.ascii_lowercase:
                        continue
                    if tzfile_hash == salt.utils.hashutils.get_hash(
                        full_path, hash_type
                    ):
                        return olson_name
    raise CommandExecutionError("Unable to determine timezone")


def _get_zone_etc_timezone():
    tzfile = "/etc/timezone"
    try:
        with salt.utils.files.fopen(tzfile, "r") as fp_:
            return salt.utils.stringutils.to_unicode(fp_.read()).strip()
    except OSError as exc:
        raise CommandExecutionError(
            f"Problem reading timezone file {tzfile}: {exc.strerror}"
        )


def _get_zone_aix():
    tzfile = "/etc/environment"
    with salt.utils.files.fopen(tzfile, "r") as fp_:
        for line in fp_:
            line = salt.utils.stringutils.to_unicode(line)
            if "TZ=" in line:
                zonepart = line.rstrip("\n").split("=")[-1]
                return zonepart.strip("'\"") or "UTC"
    raise CommandExecutionError("Unable to get timezone from " + tzfile)


def get_zone():
    """
    Get current timezone (i.e. America/Denver)

    .. versionchanged:: 2016.11.4

    .. note::

        On AIX operating systems, Posix values can also be returned
        'CST6CDT,M3.2.0/2:00:00,M11.1.0/2:00:00'

    CLI Example:

    .. code-block:: bash

        salt '*' timezone.get_zone
    """
    if salt.utils.path.which("timedatectl"):
        ret = _timedatectl()

        for line in (
            x.strip() for x in salt.utils.itertools.split(ret["stdout"], "\n")
        ):
            try:
                return re.match(r"Time ?zone:\s+(\S+)", line).group(1)
            except AttributeError:
                pass

        raise CommandExecutionError(
            "Failed to parse timedatectl output: {}\n"
            "Please file an issue with SaltStack".format(ret["stdout"])
        )

    else:
        os_family = __grains__["os_family"]
        if os_family in (
            "RedHat",
            "Suse",
        ):
            return _get_zone_sysconfig()
        if os_family in (
            "Debian",
            "Gentoo",
        ):
            return _get_zone_etc_timezone()
        if os_family in (
            "FreeBSD",
            "OpenBSD",
            "NetBSD",
            "NILinuxRT",
            "Slackware",
        ):
            return _get_zone_etc_localtime()
        if os_family in ("Solaris",):
            return _get_zone_solaris()
        if os_family in ("AIX",):
            return _get_zone_aix()
    raise CommandExecutionError("Unable to get timezone")


def get_zonecode():
    """
    Get current timezone (i.e. PST, MDT, etc)

    CLI Example:

    .. code-block:: bash

        salt '*' timezone.get_zonecode
    """
    return __salt__["cmd.run"](["date", "+%Z"], python_shell=False)


def get_offset():
    """
    Get current numeric timezone offset from UTC (i.e. -0700)

    CLI Example:

    .. code-block:: bash

        salt '*' timezone.get_offset
    """
    if "AIX" not in __grains__["os_family"]:
        return __salt__["cmd.run"](["date", "+%z"], python_shell=False)

    salt_path = "/opt/salt/bin/date"

    if not os.path.exists(salt_path):
        return f"date in salt binaries does not exist: {salt_path}"

    return __salt__["cmd.run"]([salt_path, "+%z"], python_shell=False)


def set_zone(timezone):
    """
    Unlinks, then symlinks /etc/localtime to the set timezone.

    The timezone is crucial to several system processes, each of which SHOULD
    be restarted (for instance, whatever you system uses as its cron and
    syslog daemons). This will not be automagically done and must be done
    manually!

    CLI Example:

    .. code-block:: bash

        salt '*' timezone.set_zone 'America/Denver'

    .. versionchanged:: 2016.11.4

    .. note::

        On AIX operating systems, Posix values are also allowed, see below

    .. code-block:: bash

        salt '*' timezone.set_zone 'CST6CDT,M3.2.0/2:00:00,M11.1.0/2:00:00'

    """
    if salt.utils.path.which("timedatectl"):
        try:
            __salt__["cmd.run"](f"timedatectl set-timezone {timezone}")
        except CommandExecutionError:
            pass

    if "Solaris" in __grains__["os_family"] or "AIX" in __grains__["os_family"]:
        zonepath = f"/usr/share/lib/zoneinfo/{timezone}"
    else:
        zonepath = f"/usr/share/zoneinfo/{timezone}"

    if not os.path.exists(zonepath) and "AIX" not in __grains__["os_family"]:
        return f"Zone does not exist: {zonepath}"

    tzfile = _get_localtime_path()
    if os.path.exists(tzfile):
        os.unlink(tzfile)

    if "Solaris" in __grains__["os_family"]:
        __salt__["file.sed"]("/etc/default/init", "^TZ=.*", f"TZ={timezone}")
    elif "AIX" in __grains__["os_family"]:
        # timezone could be Olson or Posix
        curtzstring = get_zone()
        cmd = ["chtz", timezone]
        result = __salt__["cmd.retcode"](cmd, python_shell=False)
        if result == 0:
            return True

        # restore orig timezone, since AIX chtz failure sets UTC
        cmd = ["chtz", curtzstring]
        __salt__["cmd.retcode"](cmd, python_shell=False)
        return False
    else:
        os.symlink(zonepath, tzfile)

    if "RedHat" in __grains__["os_family"]:
        __salt__["file.sed"]("/etc/sysconfig/clock", "^ZONE=.*", f'ZONE="{timezone}"')
    elif "Suse" in __grains__["os_family"]:
        __salt__["file.sed"](
            "/etc/sysconfig/clock", "^TIMEZONE=.*", f'TIMEZONE="{timezone}"'
        )
    elif "Debian" in __grains__["os_family"] or "Gentoo" in __grains__["os_family"]:
        with salt.utils.files.fopen("/etc/timezone", "w") as ofh:
            ofh.write(salt.utils.stringutils.to_str(timezone).strip())
            ofh.write("\n")

    return True


def zone_compare(timezone):
    """
    Compares the given timezone name with the system timezone name determined
    by timezone.get_zone. Returns True if names match, and False if not.
    Mostly useful for running state checks.

    .. versionchanged:: 2016.3.0

    .. note::

        On Solaris-like operating systems only a string comparison is done.

    .. versionchanged:: 2016.11.4

    .. note::

        On AIX operating systems only a string comparison is done.

    CLI Example:

    .. code-block:: bash

        salt '*' timezone.zone_compare 'America/Denver'
    """
    try:
        return timezone == get_zone()
    except CommandExecutionError:
        tzfile = _get_localtime_path()
        zonepath = _get_zone_file(timezone)
        try:
            return filecmp.cmp(tzfile, zonepath, shallow=False)
        except OSError as exc:
            problematic_file = exc.filename
            if problematic_file == zonepath:
                raise SaltInvocationError(f'Can\'t find a local timezone "{timezone}"')
            elif problematic_file == tzfile:
                raise CommandExecutionError(
                    "Failed to read {} to determine current timezone: {}".format(
                        tzfile, exc.strerror
                    )
                )
            raise


def _get_localtime_path():
    if (
        "NILinuxRT" in __grains__["os_family"]
        and "nilrt" in __grains__["lsb_distrib_id"]
    ):
        return "/etc/natinst/share/localtime"
    return "/etc/localtime"


def _get_zone_file(timezone):
    return f"/usr/share/zoneinfo/{timezone}"


def get_hwclock():
    """
    Get current hardware clock setting (UTC or localtime)

    CLI Example:

    .. code-block:: bash

        salt '*' timezone.get_hwclock
    """
    if salt.utils.path.which("timedatectl"):
        ret = _timedatectl()
        for line in (x.strip() for x in ret["stdout"].splitlines()):
            if "rtc in local tz" in line.lower():
                try:
                    if line.split(":")[-1].strip().lower() == "yes":
                        return "localtime"
                    else:
                        return "UTC"
                except IndexError:
                    pass

        raise CommandExecutionError(
            "Failed to parse timedatectl output: {}\n"
            "Please file an issue with SaltStack".format(ret["stdout"])
        )

    else:
        os_family = __grains__["os_family"]
        for family in ("RedHat", "Suse", "NILinuxRT"):
            if family in os_family:
                return _get_adjtime_timezone()

        if "Debian" in __grains__["os_family"]:
            # Original way to look up hwclock on Debian-based systems
            try:
                with salt.utils.files.fopen("/etc/default/rcS", "r") as fp_:
                    for line in fp_:
                        line = salt.utils.stringutils.to_unicode(line)
                        if re.match(r"^\s*#", line):
                            continue
                        if "UTC=" in line:
                            is_utc = line.rstrip("\n").split("=")[-1].lower()
                            if is_utc == "yes":
                                return "UTC"
                            else:
                                return "localtime"
            except OSError as exc:
                pass
            # Since Wheezy
            return _get_adjtime_timezone()

        if "Gentoo" in __grains__["os_family"]:
            if not os.path.exists("/etc/adjtime"):
                offset_file = "/etc/conf.d/hwclock"
                try:
                    with salt.utils.files.fopen(offset_file, "r") as fp_:
                        for line in fp_:
                            line = salt.utils.stringutils.to_unicode(line)
                            if line.startswith("clock="):
                                line = line.rstrip("\n")
                                line = line.split("=")[-1].strip("'\"")
                                if line == "UTC":
                                    return line
                                if line == "local":
                                    return "LOCAL"
                        raise CommandExecutionError(
                            f"Correct offset value not found in {offset_file}"
                        )
                except OSError as exc:
                    raise CommandExecutionError(
                        "Problem reading offset file {}: {}".format(
                            offset_file, exc.strerror
                        )
                    )
            return _get_adjtime_timezone()

        if "Solaris" in __grains__["os_family"]:
            offset_file = "/etc/rtc_config"
            try:
                with salt.utils.files.fopen(offset_file, "r") as fp_:
                    for line in fp_:
                        line = salt.utils.stringutils.to_unicode(line)
                        if line.startswith("zone_info=GMT"):
                            return "UTC"
                    return "localtime"
            except OSError as exc:
                if exc.errno == errno.ENOENT:
                    # offset file does not exist
                    return "UTC"
                raise CommandExecutionError(
                    "Problem reading offset file {}: {}".format(
                        offset_file, exc.strerror
                    )
                )

        if "AIX" in __grains__["os_family"]:
            offset_file = "/etc/environment"
            try:
                with salt.utils.files.fopen(offset_file, "r") as fp_:
                    for line in fp_:
                        line = salt.utils.stringutils.to_unicode(line)
                        if line.startswith("TZ=UTC"):
                            return "UTC"
                    return "localtime"
            except OSError as exc:
                if exc.errno == errno.ENOENT:
                    # offset file does not exist
                    return "UTC"
                raise CommandExecutionError(
                    "Problem reading offset file {}: {}".format(
                        offset_file, exc.strerror
                    )
                )

        if "Slackware" in __grains__["os_family"]:
            if not os.path.exists("/etc/adjtime"):
                offset_file = "/etc/hardwareclock"
                try:
                    with salt.utils.files.fopen(offset_file, "r") as fp_:
                        for line in fp_:
                            line = salt.utils.stringutils.to_unicode(line)
                            if line.startswith("UTC"):
                                return "UTC"
                        return "localtime"
                except OSError as exc:
                    if exc.errno == errno.ENOENT:
                        return "UTC"
            return _get_adjtime_timezone()


def set_hwclock(clock):
    """
    Sets the hardware clock to be either UTC or localtime

    CLI Example:

    .. code-block:: bash

        salt '*' timezone.set_hwclock UTC
    """
    if salt.utils.path.which("timedatectl"):
        cmd = [
            "timedatectl",
            "set-local-rtc",
            "true" if clock == "localtime" else "false",
        ]
        return __salt__["cmd.retcode"](cmd, python_shell=False) == 0
    else:
        os_family = __grains__["os_family"]
        if os_family in ("AIX", "NILinuxRT"):
            if clock.lower() != "utc":
                raise SaltInvocationError("UTC is the only permitted value")
            return True

        timezone = get_zone()

        if "Solaris" in __grains__["os_family"]:
            if clock.lower() not in ("localtime", "utc"):
                raise SaltInvocationError(
                    "localtime and UTC are the only permitted values"
                )
            if "sparc" in __grains__["cpuarch"]:
                raise SaltInvocationError(
                    "UTC is the only choice for SPARC architecture"
                )
            cmd = ["rtc", "-z", "GMT" if clock.lower() == "utc" else timezone]
            return __salt__["cmd.retcode"](cmd, python_shell=False) == 0

        zonepath = f"/usr/share/zoneinfo/{timezone}"

        if not os.path.exists(zonepath):
            raise CommandExecutionError(f"Zone '{zonepath}' does not exist")

        os.unlink("/etc/localtime")
        os.symlink(zonepath, "/etc/localtime")

        if "Arch" in __grains__["os_family"]:
            cmd = [
                "timezonectl",
                "set-local-rtc",
                "true" if clock == "localtime" else "false",
            ]
            return __salt__["cmd.retcode"](cmd, python_shell=False) == 0
        elif "RedHat" in __grains__["os_family"]:
            __salt__["file.sed"](
                "/etc/sysconfig/clock", "^ZONE=.*", f'ZONE="{timezone}"'
            )
        elif "Suse" in __grains__["os_family"]:
            __salt__["file.sed"](
                "/etc/sysconfig/clock",
                "^TIMEZONE=.*",
                f'TIMEZONE="{timezone}"',
            )
        elif "Debian" in __grains__["os_family"]:
            if clock == "UTC":
                __salt__["file.sed"]("/etc/default/rcS", "^UTC=.*", "UTC=yes")
            elif clock == "localtime":
                __salt__["file.sed"]("/etc/default/rcS", "^UTC=.*", "UTC=no")
        elif "Gentoo" in __grains__["os_family"]:
            if clock not in ("UTC", "localtime"):
                raise SaltInvocationError("Only 'UTC' and 'localtime' are allowed")
            if clock == "localtime":
                clock = "local"
            __salt__["file.sed"]("/etc/conf.d/hwclock", "^clock=.*", f'clock="{clock}"')
        elif "Slackware" in os_family:
            if clock not in ("UTC", "localtime"):
                raise SaltInvocationError("Only 'UTC' and 'localtime' are allowed")
            __salt__["file.sed"]("/etc/hardwareclock", "^(UTC|localtime)", f"{clock}")

    return True
